"""Helper to access Exif info in images."""

from __future__ import annotations

import numbers
from collections.abc import Callable
from collections.abc import Mapping
from contextlib import suppress
from datetime import datetime
from fractions import Fraction
from functools import wraps
from pathlib import Path
from typing import Any
from typing import TYPE_CHECKING
from typing import TypeAlias
from typing import TypeVar

import PIL.Image

from ._compat import ExifTags
from ._compat import UnidentifiedImageError
from .image_info import TiffOrientation


if TYPE_CHECKING:
    from _typeshed import SupportsRead
    from typing import Literal


def _combine_make(make: str | None, model: str | None) -> str:
    make = make or ""
    model = model or ""
    if make and model.startswith(make):
        return model
    return f"{make} {model}".strip()


# Interpretation of the Exif Flash tag value
#
# See: https://www.awaresystems.be/imaging/tiff/tifftags/privateifd/exif/flash.html
#
# Code copied from
# https://github.com/ianare/exif-py/blob/51d5c5adf638219632dd755c6b7a4ce2535ada62/exifread/tags/exif.py#L318-L341
#
_EXIF_FLASH_VALUES = {
    0: "Flash did not fire",
    1: "Flash fired",
    5: "Strobe return light not detected",
    7: "Strobe return light detected",
    9: "Flash fired, compulsory flash mode",
    13: "Flash fired, compulsory flash mode, return light not detected",
    15: "Flash fired, compulsory flash mode, return light detected",
    16: "Flash did not fire, compulsory flash mode",
    24: "Flash did not fire, auto mode",
    25: "Flash fired, auto mode",
    29: "Flash fired, auto mode, return light not detected",
    31: "Flash fired, auto mode, return light detected",
    32: "No flash function",
    65: "Flash fired, red-eye reduction mode",
    69: "Flash fired, red-eye reduction mode, return light not detected",
    71: "Flash fired, red-eye reduction mode, return light detected",
    73: "Flash fired, compulsory flash mode, red-eye reduction mode",
    77: (
        "Flash fired, compulsory flash mode, red-eye reduction mode, "
        "return light not detected"
    ),
    79: (
        "Flash fired, compulsory flash mode, red-eye reduction mode, "
        "return light detected"
    ),
    89: "Flash fired, auto mode, red-eye reduction mode",
    93: "Flash fired, auto mode, return light not detected, red-eye reduction mode",
    95: "Flash fired, auto mode, return light detected, red-eye reduction mode",
}


def _to_flash_description(value: int) -> str:
    desc = _EXIF_FLASH_VALUES.get(value)
    if desc is None:
        desc = f"{_EXIF_FLASH_VALUES[int(value) & 1]} ({value})"
    return desc


def _to_string(value: str) -> str:
    # XXX: By spec, strings in EXIF tags are in ASCII, however some tools
    # that handle EXIF tags support UTF-8.
    # PIL seems to return strings decoded as iso-8859-1, which is rarely, if ever,
    # right.  Attempt re-decoding as UTF-8.
    if not isinstance(value, str):
        raise ValueError(f"Value {value!r} is not a string")
    try:
        return value.encode("iso-8859-1").decode("utf-8")
    except UnicodeDecodeError:
        return value


# NB: Older versions of Pillow return (numerator, denominator) tuples
# for EXIF rational numbers.  New versions return a Fraction instance.
ExifRational: TypeAlias = numbers.Rational | tuple[int, int]
ExifReal: TypeAlias = numbers.Real | tuple[int, int]


def _to_rational(value: ExifRational) -> numbers.Rational:
    # NB: Older versions of Pillow return (numerator, denominator) tuples
    # for EXIF rational numbers.  New versions return a Fraction instance.
    if isinstance(value, numbers.Rational):
        return value
    if isinstance(value, tuple) and len(value) == 2:
        return Fraction(*value)
    raise ValueError(f"Can not convert {value!r} to Rational")


def _to_float(value: ExifReal) -> float:
    if not isinstance(value, numbers.Real):
        value = _to_rational(value)
    return float(value)


def _to_focal_length(value: ExifReal) -> str:
    return f"{_to_float(value):g}mm"


def _to_degrees(
    coords: tuple[ExifReal, ExifReal, ExifReal], hemisphere: Literal["E", "W", "N", "S"]
) -> float:
    degrees, minutes, seconds = map(_to_float, coords)
    degrees = degrees + minutes / 60 + seconds / 3600
    if hemisphere in {"S", "W"}:
        degrees = -degrees
    return degrees


def _to_altitude(altitude: ExifReal, altitude_ref: Literal[b"\x00", b"\x01"]) -> float:
    value = _to_float(altitude)
    if altitude_ref == b"\x01":
        value = -value
    return value


_T = TypeVar("_T")


def _default_none(wrapped: Callable[[EXIFInfo], _T]) -> Callable[[EXIFInfo], _T | None]:
    """Return ``None`` if wrapped getter raises a ``LookupError``.

    This is a decorator intended for use on property getters for the EXIFInfo class.

    If the wrapped getter raises a ``LookupError`` (as might happen if it tries to
    access a non-existent value in one of the EXIF tables, the wrapper will return
    ``None`` rather than propagating the exception.

    """

    @wraps(wrapped)
    def wrapper(self: EXIFInfo) -> _T | None:
        try:
            return wrapped(self)
        except LookupError:
            return None

    return wrapper


class EXIFInfo:
    """Adapt Exif tags to more user-friendly values.

    This is an adapter that wraps a ``PIL.Image.Exif`` instance to make access to
    certain Exif tags more user-friendly.

    """

    def __init__(self, exif: PIL.Image.Exif):
        self._exif = exif

    def __bool__(self) -> bool:
        """True if any Exif data exists."""
        return bool(self._exif)

    def to_dict(self) -> dict[str, str | float | tuple[float, float] | None]:
        """Return a dict containing the values of all known Exif tags."""
        rv = {}
        for key, value in self.__class__.__dict__.items():
            if key[:1] != "_" and isinstance(value, property):
                rv[key] = getattr(self, key)
        return rv

    @property
    def _ifd0(self) -> Mapping[int, Any]:
        """The main "Image File Directory" (IFD0).

        This mapping contains the basic Exif tags applying to the main image.  Keys are
        the Exif tag number, values are typing strings, ints, floats, or rationals.

        References
        ----------

        - https://www.media.mit.edu/pia/Research/deepview/exif.html#ExifTags
        - https://www.awaresystems.be/imaging/tiff/tifftags/baseline.html

        """
        return self._exif

    @property
    def _exif_ifd(self) -> Mapping[int, Any]:
        """The Exif SubIFD.

        - https://www.awaresystems.be/imaging/tiff/tifftags/privateifd/exif.html
        """
        return self._exif.get_ifd(ExifTags.IFD.Exif)  # type: ignore[no-any-return]

    @property
    def _gpsinfo_ifd(self) -> Mapping[int, Any]:
        """The GPS IFD

        - https://www.awaresystems.be/imaging/tiff/tifftags/privateifd/gps.html
        """
        # On older Pillow versions, get_ifd(GPSinfo) returns None.
        # Prior to somewhere around Pillow 8.2.0, the GPS IFD was accessible at
        # the top level. Try that first.
        #
        # https://pillow.readthedocs.io/en/stable/releasenotes/8.2.0.html#image-getexif-exif-and-gps-ifd
        gps_ifd = self._exif.get(ExifTags.IFD.GPSInfo)
        if isinstance(gps_ifd, dict):
            return gps_ifd
        return self._exif.get_ifd(ExifTags.IFD.GPSInfo)  # type: ignore[no-any-return]

    @property
    @_default_none
    def artist(self) -> str:
        return _to_string(self._ifd0[ExifTags.Base.Artist])

    @property
    @_default_none
    def copyright(self) -> str:
        return _to_string(self._ifd0[ExifTags.Base.Copyright])

    @property
    @_default_none
    def camera_make(self) -> str:
        return _to_string(self._ifd0[ExifTags.Base.Make])

    @property
    @_default_none
    def camera_model(self) -> str:
        return _to_string(self._ifd0[ExifTags.Base.Model])

    @property
    def camera(self) -> str:
        return _combine_make(self.camera_make, self.camera_model)

    @property
    @_default_none
    def lens_make(self) -> str:
        return _to_string(self._exif_ifd[ExifTags.Base.LensMake])

    @property
    @_default_none
    def lens_model(self) -> str:
        return _to_string(self._exif_ifd[ExifTags.Base.LensModel])

    @property
    def lens(self) -> str:
        return _combine_make(self.lens_make, self.lens_model)

    @property
    @_default_none
    def aperture(self) -> float:
        return round(_to_float(self._exif_ifd[ExifTags.Base.ApertureValue]), 4)

    @property
    @_default_none
    def f_num(self) -> float:
        return round(_to_float(self._exif_ifd[ExifTags.Base.FNumber]), 4)

    @property
    @_default_none
    def f(self) -> str:
        value = _to_float(self._exif_ifd[ExifTags.Base.FNumber])
        return f"ƒ/{value:g}"

    @property
    @_default_none
    def exposure_time(self) -> str:
        value = _to_rational(self._exif_ifd[ExifTags.Base.ExposureTime])
        return f"{value.numerator}/{value.denominator}"

    @property
    @_default_none
    def shutter_speed(self) -> str:
        value = _to_float(self._exif_ifd[ExifTags.Base.ShutterSpeedValue])
        return f"1/{2**value:.0f}"

    @property
    @_default_none
    def focal_length(self) -> str:
        return _to_focal_length(self._exif_ifd[ExifTags.Base.FocalLength])

    @property
    @_default_none
    def focal_length_35mm(self) -> str:
        return _to_focal_length(self._exif_ifd[ExifTags.Base.FocalLengthIn35mmFilm])

    @property
    @_default_none
    def flash_info(self) -> str:
        return _to_flash_description(self._exif_ifd[ExifTags.Base.Flash])

    @property
    @_default_none
    def iso(self) -> float:
        return _to_float(self._exif_ifd[ExifTags.Base.ISOSpeedRatings])

    @property
    def created_at(self) -> datetime | None:
        date_tags = (
            # XXX: GPSDateStamp includes just the date
            # https://www.awaresystems.be/imaging/tiff/tifftags/privateifd/gps/gpsdatestamp.html
            (self._gpsinfo_ifd, ExifTags.GPS.GPSDateStamp),
            # XXX: DateTimeOriginal is an EXIF tag, not and IFD0 tag
            (self._ifd0, ExifTags.Base.DateTimeOriginal),
            (self._exif_ifd, ExifTags.Base.DateTimeOriginal),
            (self._exif_ifd, ExifTags.Base.DateTimeDigitized),
            (self._ifd0, ExifTags.Base.DateTime),
        )
        for ifd, tag in date_tags:
            with suppress(LookupError, ValueError):
                return datetime.strptime(ifd[tag], "%Y:%m:%d %H:%M:%S")
        return None

    @property
    @_default_none
    def longitude(self) -> float:
        gpsinfo_ifd = self._gpsinfo_ifd
        return _to_degrees(
            gpsinfo_ifd[ExifTags.GPS.GPSLongitude],
            gpsinfo_ifd[ExifTags.GPS.GPSLongitudeRef],
        )

    @property
    @_default_none
    def latitude(self) -> float:
        gpsinfo_ifd = self._gpsinfo_ifd
        return _to_degrees(
            gpsinfo_ifd[ExifTags.GPS.GPSLatitude],
            gpsinfo_ifd[ExifTags.GPS.GPSLatitudeRef],
        )

    @property
    @_default_none
    def altitude(self) -> float:
        gpsinfo_ifd = self._gpsinfo_ifd
        value = _to_float(gpsinfo_ifd[ExifTags.GPS.GPSAltitude])
        ref = gpsinfo_ifd.get(ExifTags.GPS.GPSAltitudeRef)
        if ref == b"\x01":
            value = -value
        return value

    @property
    def location(self) -> tuple[float, float] | None:
        lat = self.latitude
        long = self.longitude
        if lat is not None and long is not None:
            return (lat, long)
        return None

    @property
    @_default_none
    def documentname(self) -> str:
        return _to_string(self._ifd0[ExifTags.Base.DocumentName])

    @property
    @_default_none
    def description(self) -> str:
        return _to_string(self._ifd0[ExifTags.Base.ImageDescription])

    @property
    def is_rotated(self) -> bool:
        """Return if the image is rotated according to the Orientation header.

        The Orientation header in EXIF stores an integer value between
        1 and 8, where the values 5-8 represent "portrait" orientations
        (rotated 90deg left, right, and mirrored versions of those), i.e.,
        the image is rotated.
        """
        try:
            orientation = TiffOrientation(self._ifd0[ExifTags.Base.Orientation])
        except (LookupError, ValueError):
            return False
        return orientation.is_transposed


def read_exif(source: str | Path | SupportsRead[bytes]) -> EXIFInfo:
    """Reads exif data from an image file."""
    try:
        with PIL.Image.open(source) as image:
            exif = image.getexif()
    except UnidentifiedImageError:
        exif = PIL.Image.Exif()
    return EXIFInfo(exif)



import getpass
import os
import re
import shutil
from contextlib import contextmanager
from datetime import datetime
from functools import partial
from importlib import import_module
from subprocess import PIPE
from subprocess import run
from tempfile import TemporaryDirectory

import click
from jinja2 import Environment
from jinja2 import PackageLoader

from lektor.utils import locate_executable
from lektor.utils import slugify


pwd = import_module("pwd") if os.name != "nt" else None


_var_re = re.compile(r"@([^@]+)@")


class Generator:
    def __init__(self, base):
        self.question = 0
        self.jinja_env = Environment(
            loader=PackageLoader("lektor", os.path.join("quickstart-templates", base)),
            line_statement_prefix="%%",
            line_comment_prefix="##",
            variable_start_string="${",
            variable_end_string="}",
            block_start_string="<%",
            block_end_string="%>",
            comment_start_string="/**",
            comment_end_string="**/",
        )
        self.options = {}
        # term width in [1, 78]
        self.term_width = min(max(shutil.get_terminal_size()[0], 1), 78)
        self.e = click.secho
        self.w = partial(click.wrap_text, width=self.term_width)

    @staticmethod
    def abort(message):
        click.echo(f"Error: {message}", err=True)
        raise click.Abort()

    def prompt(self, text, default=None, info=None):
        self.question += 1
        self.e("")
        self.e(f"Step {self.question}:", fg="yellow")
        if info is not None:
            self.e(click.wrap_text(info, self.term_width, "| ", "| "))
        text = "> " + click.style(text, fg="green")

        if default is True or default is False:
            return click.confirm(text, default=default)
        return click.prompt(text, default=default, show_default=True)

    def title(self, title):
        self.e(title, fg="cyan")
        self.e("=" * len(title), fg="cyan")
        self.e("")

    def warn(self, text):
        self.e(self.w(text), fg="magenta")

    def text(self, text):
        self.e(self.w(text))

    def confirm(self, prompt):
        self.e("")
        click.confirm(prompt, default=True, abort=True, prompt_suffix=" ")

    @contextmanager
    def make_target_directory(self, path):
        here = os.path.abspath(os.getcwd())
        path = os.path.abspath(path)
        if here != path:
            try:
                os.makedirs(path)
            except OSError as e:
                self.abort(f"Could not create target folder: {e}")

        if os.path.isdir(path):
            try:
                if len(os.listdir(path)) != 0:
                    raise OSError("Directory not empty")
            except OSError as e:
                self.abort(f"Bad target folder: {e}")

        with TemporaryDirectory() as scratch:
            yield scratch

            # Use shutil.move here in case we move across a file system
            # boundary.
            for filename in os.listdir(scratch):
                shutil.move(
                    os.path.join(scratch, filename), os.path.join(path, filename)
                )

    @staticmethod
    def expand_filename(base, ctx, template_filename):
        def _repl(match):
            return ctx[match.group(1)]

        return os.path.join(base, _var_re.sub(_repl, template_filename))[:-3]

    def run(self, ctx, path):
        with self.make_target_directory(path) as scratch:
            for template in self.jinja_env.list_templates():
                if not template.endswith(".in"):
                    continue
                fn = self.expand_filename(scratch, ctx, template)
                tmpl = self.jinja_env.get_template(template)
                rv = tmpl.render(ctx).strip("\r\n")
                if rv:
                    directory = os.path.dirname(fn)
                    try:
                        os.makedirs(directory)
                    except OSError:
                        pass
                    with open(fn, "wb") as f:
                        f.write((rv + "\n").encode("utf-8"))


def get_default_author() -> str:
    """Attempt to guess an the name of the current user."""
    if pwd is not None:
        try:
            pw_gecos = pwd.getpwuid(os.getuid()).pw_gecos
        except KeyError:
            pass
        else:
            full_name = pw_gecos.split(",", 1)[0].strip()
            if full_name:
                return full_name

    return getpass.getuser()


def get_default_author_email() -> str | None:
    """Attempt to guess an email address for the current user.

    May return an empty string if not reasonable guess can be made.
    """
    git = locate_executable("git")
    if git:
        proc = run(
            (git, "config", "user.email"), stdout=PIPE, errors="strict", check=False
        )
        if proc.returncode == 0:
            return proc.stdout.strip()

    email = os.environ.get("EMAIL", "").strip()
    if email:
        return email
    # We could fall back to f"{getpass.getuser()}@{socket.getfqdn()}",
    # but it is probably better just to go with no default in that
    # case.
    return None


def project_quickstart(defaults=None):
    if not defaults:
        defaults = {}

    g = Generator("project")

    g.title("Lektor Quickstart")
    g.text(
        "This wizard will generate a new basic project with some sensible "
        "defaults for getting started quickly.  We just need to go through "
        "a few questions so that the project is set up correctly for you."
    )

    name = defaults.get("name")
    if name is None:
        name = g.prompt(
            "Project Name",
            None,
            "A project needs a name.  The name is primarily used for the admin "
            "UI and some other places to refer to your project to not get "
            "confused if multiple projects exist.  You can change this at "
            "any later point.",
        )

    author_name = g.prompt(
        "Author Name",
        get_default_author(),
        "Your name.  This is used in a few places in the default template "
        "to refer to in the default copyright messages.",
    )

    path = defaults.get("path")
    if path is None:
        default_project_path = os.path.join(os.getcwd(), name)
        path = g.prompt(
            "Project Path",
            default_project_path,
            "This is the path where the project will be located.  You can "
            "move a project around later if you do not like the path.  If "
            "you provide a relative path it will be relative to the working "
            "directory.",
        )
        path = os.path.expanduser(path)

    with_blog = g.prompt(
        "Add Basic Blog",
        True,
        "Do you want to generate a basic blog module?  If you enable this "
        "the models for a very basic blog will be generated.",
    )

    g.confirm("That's all. Create project?")

    g.run(
        {
            "project_name": name,
            "project_slug": slugify(name),
            "project_path": path,
            "with_blog": with_blog,
            "this_year": datetime.utcnow().year,
            "today": datetime.utcnow().strftime("%Y-%m-%d"),
            "author_name": author_name,
        },
        path,
    )


def plugin_quickstart(defaults=None, project=None):
    if defaults is None:
        defaults = {}

    g = Generator("plugin")

    plugin_name = defaults.get("plugin_name")
    if plugin_name is None:
        plugin_name = g.prompt(
            "Plugin Name",
            default=None,
            info="This is the human readable name for this plugin",
        )

    plugin_id = plugin_name.lower()
    plugin_id = plugin_id.removeprefix("lektor")
    plugin_id = plugin_id.removesuffix("plugin")
    plugin_id = slugify(plugin_id)

    path = defaults.get("path")
    if path is None:
        if project is not None:
            default_path = os.path.join(project.tree, "packages", plugin_id)
        else:
            if len(os.listdir(".")) == 0:
                default_path = os.getcwd()
            else:
                default_path = os.path.join(os.getcwd(), plugin_id)
        path = g.prompt(
            "Plugin Path",
            default_path,
            "The place where you want to initialize the plugin",
        )

    author_name = g.prompt(
        "Author Name",
        get_default_author(),
        "Your name as it will be embedded in the plugin metadata.",
    )

    author_email = g.prompt(
        "Author E-Mail",
        get_default_author_email(),
        "Your e-mail address for the plugin info.",
    )

    g.confirm("Create Plugin?")

    g.run(
        {
            "plugin_name": plugin_name,
            "plugin_id": plugin_id,
            "plugin_class": plugin_id.title().replace("-", "") + "Plugin",
            "plugin_module": "lektor_" + plugin_id.replace("-", "_"),
            "author_name": author_name,
            "author_email": author_email,
        },
        path,
    )


def theme_quickstart(defaults=None, project=None):
    if defaults is None:
        defaults = {}

    g = Generator("theme")

    theme_name = defaults.get("theme_name")
    if theme_name is None:
        theme_name = g.prompt(
            "Theme Name",
            default=None,
            info="This is the human readable name for this theme",
        )

    theme_id = theme_name.lower()
    if theme_id != "lektor" and theme_id.startswith("lektor"):
        theme_id = theme_id[6:].strip()
    if theme_id != "theme" and theme_id.startswith("theme"):
        theme_id = theme_id[5:]
    if theme_id != "theme" and theme_id.endswith("theme"):
        theme_id = theme_id[:-5]
    theme_id = slugify(theme_id)

    path = defaults.get("path")
    if path is None:
        if project is not None:
            default_path = os.path.join(
                project.tree, "themes", f"lektor-theme-{theme_id}"
            )
        else:
            if len(os.listdir(".")) == 0:
                default_path = os.getcwd()
            else:
                default_path = os.path.join(os.getcwd(), theme_id)
        path = g.prompt(
            "Theme Path",
            default_path,
            "The place where you want to initialize the theme",
        )

    author_name = g.prompt(
        "Author Name",
        get_default_author(),
        "Your name as it will be embedded in the theme metadata.",
    )

    author_email = g.prompt(
        "Author E-Mail",
        get_default_author_email(),
        "Your e-mail address for the theme info.",
    )

    g.confirm("Create Theme?")

    g.run(
        {
            "theme_name": theme_name,
            "theme_id": theme_id,
            "author_name": author_name,
            "author_email": author_email,
        },
        path,
    )

    # symlink
    theme_dir = os.getcwd()
    example_themes = os.path.join(path, "example-site/themes")
    os.makedirs(example_themes)
    os.chdir(example_themes)
    try:
        os.symlink(
            f"../../../lektor-theme-{theme_id}",
            f"lektor-theme-{theme_id}",
        )
    except OSError as exc:
        # Windows, by default, only allows members of the "Administrators" group
        # to create symlinks. For users who are not allowed to create symlinks,
        # error Code 1314 - "A required privilege is not held by the client"
        # is raised.
        if getattr(exc, "winerror", None) != 1314:
            raise
        g.warn(
            "Could not automatically make a symlink to have your example-site"
            "easily pick up your theme."
        )
    os.chdir(theme_dir)

    # Sample image
    os.makedirs(os.path.join(path, "images"))
    source_image_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "quickstart-templates/theme/images/homepage.png",
    )
    destination_image_path = os.path.join(path, "images/homepage.png")
    with open(source_image_path, "rb") as f:
        image = f.read()
    with open(destination_image_path, "wb") as f:
        f.write(image)



from __future__ import annotations

from typing import TYPE_CHECKING
from wsgiref.util import shift_path_info

from flask import Flask
from flask import request

from lektor.admin.context import LektorApp
from lektor.admin.context import LektorInfo
from lektor.admin.modules import api
from lektor.admin.modules import dash
from lektor.admin.modules import livereload
from lektor.admin.modules import serve
from lektor.environment import Environment


if TYPE_CHECKING:
    from _typeshed import StrPath
    from _typeshed.wsgi import WSGIApplication


def _common_configuration(app: Flask, debug: bool = False) -> None:
    app.debug = debug
    app.config["PROPAGATE_EXCEPTIONS"] = True


def make_app(
    env: Environment,
    debug: bool = False,
    output_path: StrPath | None = None,
    ui_lang: str = "en",
    verbosity: int = 0,
    extra_flags: dict[str, str] | None = None,
    *,
    admin_path: str = "/admin",
    static_folder: StrPath | None = "static",  # testing
) -> LektorApp:
    if output_path is None:
        output_path = env.project.get_output_path()

    lektor_info = LektorInfo(env, output_path, verbosity, extra_flags)

    # The top-level app has a route that matches anything
    # ("/<path:path>" in the serve blueprint).  That means that if
    # there is another route whose doesn't match based on request
    # method, the serve view will take over and try to serve it.  To
    # prevent this from happening for the paths under /admin, we
    # structure them as a separate flask app.
    admin_app = LektorApp(lektor_info)
    _common_configuration(admin_app, debug=debug)
    admin_app.config["lektor.ui_lang"] = ui_lang
    admin_app.register_blueprint(dash.bp, url_prefix="/")
    admin_app.register_blueprint(api.bp, url_prefix="/api")

    # Serve static files from top-level app
    app = LektorApp(
        lektor_info, static_url_path=f"{admin_path}/static", static_folder=static_folder
    )
    _common_configuration(app, debug=debug)
    app.register_blueprint(livereload.bp, url_prefix="/__reload__")
    app.register_blueprint(serve.bp)

    # Pass requests for /admin/... to the admin app
    @app.route(f"{admin_path}/", defaults={"page": ""})
    @app.route(f"{admin_path}/<path:page>", methods=["GET", "POST", "PUT"])
    def admin_view(page: str) -> WSGIApplication:
        environ = request.environ
        # Save top-level SCRIPT_NAME (used by dash)
        environ["lektor.site_root"] = request.root_path
        while environ.get("PATH_INFO", "") != f"/{page}":
            assert environ["PATH_INFO"]
            shift_path_info(request.environ)
        return admin_app.wsgi_app

    # Add rule to construct URL to /admin/edit
    app.add_url_rule(f"{admin_path}/edit", "url.edit", build_only=True)

    return app


WebAdmin = WebUI = make_app



import re
import threading

import pytest
from markupsafe import Markup

import lektor.markdown
from lektor.context import Context
from lektor.markdown import controller_class
from lektor.markdown import get_controller
from lektor.markdown import make_markdown
from lektor.markdown import Markdown
from lektor.markdown import markdown_to_html
from lektor.markdown import MISTUNE_VERSION
from lektor.markdown.controller import get_renderer_context
from lektor.markdown.controller import RendererContext
from lektor.markdown.controller import RendererHelper
from lektor.markdown.controller import require_ctx
from lektor.markdown.controller import UnknownPluginError
from lektor.pluginsystem import Plugin


@pytest.fixture
def record_path():
    return "/extra"


@pytest.fixture
def record_alt():
    return "en"


@pytest.fixture
def record(pad, record_path, record_alt):
    return pad.get(record_path, alt=record_alt)


@pytest.fixture
def resolve_links():
    return None


@pytest.fixture
def field_options(resolve_links):
    """Field options"""
    return {
        "label": "Test Markdown Field",
        "type": "markdown",
        "resolve_links": resolve_links,
    }


@pytest.fixture
def base_url(record):
    return None


@pytest.fixture
def context(pad, record, base_url):
    if base_url is None:
        base_url = record.url_path
    with Context(pad=pad) as ctx:
        with ctx.changed_base_url(base_url):
            yield ctx


@pytest.fixture
def renderer_context(context, record, field_options):
    with RendererContext(record, {}, field_options) as renderer_context:
        yield renderer_context


def test_require_ctx_raises_if_no_ctx():
    with pytest.raises(RuntimeError) as exc_info:
        require_ctx()
    assert "required for markdown rendering" in str(exc_info.value)


def _run_in_thread(target):
    thread = threading.Thread(target=target)
    thread.start()
    thread.join()


def test_renderer_context_is_thread_local(renderer_context):
    assert get_renderer_context() is renderer_context

    @_run_in_thread
    def assert_no_renderer_context():
        with pytest.raises(RuntimeError):
            get_renderer_context()


@pytest.mark.usefixtures("renderer_context")
def test_RendererHelper_options(field_options):
    helper = RendererHelper()
    assert helper.field_options == field_options


@pytest.mark.parametrize(
    "record_path, record_alt, url, base_url, resolve_links, expected",
    [
        ("/", "en", "test.jpg", None, None, "test.jpg"),
        ("/extra", "en", "missing", "/blog/", None, "../extra/missing"),
        ("/extra", "en", "a", "/blog/", None, "../extra/a/"),
        ("/extra", "en", "slash-slug", None, None, "long/path/"),
        ("/extra", "en", "slash-slug", None, "never", "slash-slug"),
        ("/", "de", "test.jpg", None, None, "../test.jpg"),
    ],
)
@pytest.mark.usefixtures("renderer_context")
def test_RendererHelper_resolve_url(url, expected):
    helper = RendererHelper()
    assert helper.resolve_url(url) == expected


@pytest.mark.parametrize("resolve_links", ["always"])
@pytest.mark.usefixtures("renderer_context")
def test_RendererHelper_resolve_url_raises_when_not_resolvable():
    helper = RendererHelper()
    with pytest.raises(RuntimeError) as exc_info:
        helper.resolve_url("missing")
    assert re.search(r"Can not resolve .*missing", str(exc_info.value))


@pytest.mark.parametrize("record, base_url", [(None, "/")])
@pytest.mark.usefixtures("renderer_context")
def test_RendererHelper_resolve_url_no_record():
    helper = RendererHelper()
    assert helper.resolve_url("foo/bar") == "foo/bar"


@pytest.mark.parametrize("resolve_links", ["always"])
@pytest.mark.parametrize("record, base_url", [(None, "/")])
@pytest.mark.usefixtures("renderer_context")
def test_RendererHelper_resolve_url_raises_if_no_record():
    helper = RendererHelper()
    with pytest.raises(RuntimeError) as exc_info:
        helper.resolve_url(".")
    assert re.search(r"\bsource object\b.*\brequired\b", str(exc_info.value))


@pytest.fixture
def markdown_controller(env):
    return controller_class(env)


def test_markdown_controller_parser_caching(markdown_controller):
    parsers = {markdown_controller.parser}

    def get_parser():
        parsers.add(markdown_controller.parser)

    get_parser()
    assert len(parsers) == 1

    _run_in_thread(get_parser)
    if MISTUNE_VERSION.startswith("2."):
        # mistune 2.x's parser is thread-safe. It need not be thread-local.
        assert len(parsers) == 1
    else:
        # mistune 0.x's parser is not thread-safe. We need one for each thread.
        assert len(parsers) == 2


@pytest.mark.parametrize("base_url", ["/BASE/"])
@pytest.mark.usefixtures("context")
def test_markdown_controller_get_cache_key(markdown_controller, base_url):
    assert markdown_controller.get_cache_key() == base_url


def test_markdown_controller_get_cache_key_no_context(markdown_controller):
    assert markdown_controller.get_cache_key() is None


def test_controller_cache(env):
    controller = get_controller(env)
    assert get_controller(env) is controller


if MISTUNE_VERSION.startswith("0."):
    plugin_url = ...
else:
    from mistune.plugins import plugin_url


def plugin_null(md):
    return md


@pytest.mark.skipif(MISTUNE_VERSION.startswith("0."), reason="mistune0")
@pytest.mark.parametrize(
    "plugin, resolved",
    [
        (plugin_null, plugin_null),  # callable is already resolved
        ("url", plugin_url),  # resolve through mistune.PLUGINS
        ("mistune.plugins:plugin_url", plugin_url),  # resolve via import
    ],
)
def test_markdown_controller_resolve_plugin(plugin, resolved, markdown_controller):
    assert markdown_controller.resolve_plugin(plugin) is resolved


@pytest.mark.skipif(MISTUNE_VERSION.startswith("0."), reason="mistune0")
@pytest.mark.parametrize("plugin", ["badplugin", "unknown_module:badplugin"])
def test_markdown_controller_resolve_plugin_unknown(plugin, markdown_controller):
    with pytest.raises(UnknownPluginError):
        markdown_controller.resolve_plugin(plugin)


@pytest.mark.skipif(MISTUNE_VERSION.startswith("0."), reason="mistune0")
def test_markdown_controller_resolve_plugin_raises_typeerror(markdown_controller):
    with pytest.raises(TypeError):
        markdown_controller.resolve_plugin(None)


@pytest.fixture
def improved_renderer():
    # pylint: disable=import-outside-toplevel
    if MISTUNE_VERSION.startswith("2."):
        from lektor.markdown.mistune2 import ImprovedRenderer
    else:
        from lektor.markdown.mistune0 import ImprovedRenderer
    return ImprovedRenderer()


@pytest.mark.parametrize(
    "link, title, text, expected",
    [
        # NB: Mistune 0.x and mistune 2.x behave differently with
        # respect to how much HTML escaping of the ``link`` the
        # renderer is expected to do.  Do not put HTML-escaping tests
        # here.  Put them in the ``test_integration`` test later in
        # this file.
        ("a", None, "text", r'<a href="a/">text</a>\Z'),
        ("missing", None, "text", r'<a href="missing">text</a>\Z'),
        ("/", "T", "text", r'<a href="../" title="T">text</a>\Z'),
        ("a&amp;b", None, "x", r'.* href="a&amp;b"'),
        ("/", "<title>", "x", r'.* title="&lt;title&gt;"'),
    ],
)
@pytest.mark.usefixtures("renderer_context")
def test_ImprovedRenderer_link(link, title, text, expected, improved_renderer):
    if MISTUNE_VERSION.startswith("2."):
        result = improved_renderer.link(link, text, title)
    else:
        result = improved_renderer.link(link, title, text)
    assert re.match(expected, result)


@pytest.mark.parametrize(
    "src, title, alt, expected",
    [
        # NB: Mistune 0.x and mistune 2.x behave differently with
        # respect to how much HTML escaping of the ``src`` the
        # renderer is expected to do.  Do not put HTML-escaping tests
        # here.  Put them in the ``test_integration`` test later in
        # this file.
        ("/test.jpg", None, "text", r'<img src="../test.jpg" alt="text"\s*/?>\Z'),
        ("/test.jpg", "T", "x", r'.* title="T"'),
        ("&amp;c.gif", None, "x", r'.* src="&amp;c.gif"'),
        ("/test.jpg", "<title>", "x", r'.* title="&lt;title&gt;"'),
        ("/test.jpg", None, "x&y", r'.* alt="x&amp;y"'),
    ],
)
@pytest.mark.usefixtures("renderer_context")
def test_ImprovedRenderer_image(src, title, alt, expected, improved_renderer):
    if MISTUNE_VERSION.startswith("2."):
        result = improved_renderer.image(src, alt, title)
    else:
        result = improved_renderer.image(src, title, alt)
    assert re.match(expected, result.rstrip())


def plugin_linkcount(md):
    """A test mistune2 plugin.

    This replaces ``[link-count]`` in text with the current link count
    wrapped in a <code> element.
    """

    def parse_linkcount(inline, _m, _state):
        nlinks = inline.renderer.lektor.meta["nlinks"]
        return "codespan", str(nlinks)

    md.inline.register_rule("link_count", r"\[link-count\]", parse_linkcount)
    index = md.inline.rules.index("ref_link2")
    md.inline.rules.insert(index, "link_count")


class LinkCounterPlugin(Plugin):
    """A test plugin to test the renderer meta system."""

    name = "LinkCounter"

    class RendererMixin:
        def link(self, *args, **kwargs):
            multiplier = int(self.lektor.field_options.get("multiplier", 1))
            self.lektor.meta["nlinks"] += multiplier
            return super().link(*args, **kwargs)

    def on_markdown_config(self, config, **kwargs):
        config.renderer_mixins.append(self.RendererMixin)

    @staticmethod
    def on_markdown_lexer_config(config, renderer, **kwargs):
        if not hasattr(config, "parser_options"):
            return  # mistune 0
        # Configure a plugin just to make sure we can
        config.parser_options.setdefault("plugins", []).append(plugin_linkcount)

    def on_markdown_meta_init(self, meta, **kwargs):
        # pylint: disable=no-self-use
        meta["nlinks"] = 0


@pytest.fixture
def link_counter_plugin(env):
    env.plugin_controller.instanciate_plugin("link-counter", LinkCounterPlugin)


class TestMarkdown:
    # pylint: disable=no-self-use

    @pytest.fixture
    def source(self):
        return "text"

    @pytest.fixture
    def markdown(self, source, record, field_options):
        return Markdown(source, record, field_options)

    @pytest.mark.parametrize("source, expected", [("x", True), ("", False)])
    def test_bool(self, markdown, source, expected):
        assert bool(markdown) is expected

    def test_record(self, markdown, record):
        assert markdown.record is record

    @pytest.mark.parametrize("record", [None])
    def test_record_none(self, markdown):
        assert markdown.record is None

    def test_record_gone_away(self, field_options, mocker):
        markdown = Markdown("text", mocker.Mock(name="record"), field_options)
        with pytest.raises(RuntimeError) as exc_info:
            _ = markdown.record
        assert "gone away" in str(exc_info.value)

    @pytest.mark.parametrize("source", ["![test](/test.jpg)"])
    def test_render_caching(self, markdown, context):
        render1 = markdown._Markdown__render()
        assert '<img src="../test.jpg"' in render1.html
        with context.changed_base_url("/"):
            render2 = markdown._Markdown__render()
            assert '<img src="test.jpg"' in render2.html
        render3 = markdown._Markdown__render()
        assert render3 is render1

    @pytest.mark.parametrize(
        "source, nlinks",
        [
            ("howdy", 0),
            ("[howdy](http://example.com/)", 1),
        ],
    )
    @pytest.mark.usefixtures("context", "link_counter_plugin")
    def test_meta(self, markdown, nlinks):
        assert markdown.meta["nlinks"] == nlinks

    @pytest.mark.usefixtures("context")
    def test_html(self, markdown):
        assert markdown.html.rstrip() == "<p>text</p>"

    @pytest.mark.usefixtures("context", "link_counter_plugin")
    def test_getitem(self, markdown):
        assert markdown["nlinks"] == 0

    @pytest.mark.parametrize("source", ["[x](/y) [link-count]"])
    @pytest.mark.usefixtures("context", "link_counter_plugin")
    @pytest.mark.skipif(MISTUNE_VERSION.startswith("0."), reason="mistune0")
    def test_linkcount_plugin(self, markdown):
        assert markdown["nlinks"] == 1
        assert re.search(r"<code>1</code>", markdown.html)

    @pytest.mark.usefixtures("context")
    def test_str(self, markdown):
        assert str(markdown).rstrip() == "<p>text</p>"

    @pytest.mark.usefixtures("context")
    def test_markup(self, markdown):
        assert markdown.__html__().rstrip() == "<p>text</p>"


def _normalize_html(output: str | Markup) -> str:
    html = str(output).strip()
    html = html.replace("&copy;", "©")
    html = re.sub(r"(<img [^>]*?)\s*/>", r"\1>", html)
    return html


@pytest.mark.parametrize(
    "source, expected",
    [
        ("text", r"<p>text</p>\Z"),
        ("&copy;", r"<p>©</p>\Z"),
        # Various link texts
        ("[text](x)", r".*<a [^>]*>text</a>"),
        ("[&copy;](x)", r".*<a [^>]*>©</a>"),
        ("[a&b](x)", r".*<a [^>]*>a&amp;b</a>"),
        ("[a&amp;b](x)", r".*<a [^>]*>a&amp;b</a>"),
        ("[a<>b](x)", r".*<a [^>]*>a&lt;&gt;b</a>"),
        ("[a<br>b](x)", r".*<a [^>]*>a<br>b</a>"),
        # Various link titles
        ('[x](x "TITLE")', r'.*<a\b.* title="TITLE">'),
        ('[x](x "©")', r'.*<a\b.* title="©">'),
        ('[x](x "&copy;")', r'.*<a\b.* title="©">'),
        ('[x](x "a&b")', r'.*<a\b.* title="a&amp;b">'),
        ('[x](x "a&amp;b")', r'.*<a\b.* title="a&amp;b">'),
        ('[x](x "a<br>b")', r'.*<a\b.* title="a&lt;br&gt;b">'),
        # Various link hrefs
        ("[x](&amp;c.html)", r'.*<a\b.* href="&amp;c.html">'),
        ("[x](&c.html)", r'.*<a\b.* href="&amp;c.html">'),
        # Various image alts
        ("![©](x)", r'.*<img\b[^>]* alt="©"'),
        ("![&copy;](x)", r'.*<img\b[^>]* alt="©"'),
        ("![<br>](x)", r'.*<img\b[^>]* alt="&lt;br&gt;"'),
        ("![&](x)", r'.*<img\b[^>]* alt="&amp;"'),
        ("![&amp;](x)", r'.*<img\b[^>]* alt="&amp;"'),
        # Various image titles
        ('![x](y "<br>")', r'.*<img\b[^>]* title="&lt;br&gt;"'),
        ('![x](y "&lt;")', r'.*<img\b[^>]* title="&lt;"'),
        # Various image srcs
        ("![](x&y.gif)", r'.*<img\b[^>]* src="x&amp;y.gif"'),
        ("![](x&amp;y.gif)", r'.*<img\b[^>]* src="x&amp;y.gif"'),
        # FIXME: which of these is right?
        ("![](x<br>y.gif)", r'.*<img\b[^>]* src="x(&lt;br&gt;|%3Cbr%3E)y.gif"'),
        # Autolink
        ("autolink: <http://example.org>", r'.*<a href="http://example\.org"'),
        #################
        # Extra features enabled by default in mistune 0.8.4, but which require
        # plugins to be enabled in mistune 2
        ("strikethrough: ~~stricken~~ foo", r".*<del>stricken</del> foo"),
        (
            "footnotes[^note]\n\n[^note]: comment",
            r'(?s).*<a href="#fn-([^"]+)">.*<li id="fn-\1">',
        ),
        ("| Table Header |\n| -------|\n| Cell |\n", r"(?s).*<tr>"),
        ("raw URL - http://example.net/ foo", r'.*<a href="http://example\.net/"'),
    ],
)
@pytest.mark.usefixtures("context")
def test_integration(record, field_options, source, expected):
    rendered = _normalize_html(Markdown(source, record, field_options))
    assert re.match(expected, rendered)


@pytest.mark.usefixtures("context", "link_counter_plugin")
def test_field_options_integration(record, field_options):
    # tests that we can pass field options to a custom renderer mixin
    field_options = dict(field_options, multiplier="11")
    two_links = "[one](x.html), [two](y.html)"
    markdown = Markdown(two_links, record, field_options)
    assert markdown.meta["nlinks"] == 22


################################################################
#
# Tests for the bits that are provided for backward compatibility
# with plugins written for Lektor < 3.4.
#
def test_deprecated_ImprovedRenderer(improved_renderer):
    with pytest.deprecated_call(
        match=r"lektor\.markdown\.ImprovedRenderer\b.* deprecated"
    ) as warnings:
        # pylint: disable-next=import-outside-toplevel,no-name-in-module
        from lektor.markdown import ImprovedRenderer
    assert all(w.filename == __file__ for w in warnings)
    assert isinstance(improved_renderer, ImprovedRenderer)


@pytest.mark.skipif(not MISTUNE_VERSION.startswith("0."), reason="not mistune0")
@pytest.mark.usefixtures("renderer_context")
def test_deprecated_ImprovedRenderer_record(record):
    with pytest.deprecated_call():
        # pylint: disable-next=import-outside-toplevel,no-name-in-module
        from lektor.markdown import ImprovedRenderer
    improved_renderer = ImprovedRenderer()

    with pytest.deprecated_call(
        match=r"Use .*Renderer\.lektor\.record instead"
    ) as warnings:
        assert improved_renderer.record is record
    assert all(w.filename == __file__ for w in warnings)


@pytest.mark.skipif(not MISTUNE_VERSION.startswith("0."), reason="not mistune0")
def test_deprecated_ImprovedRenderer_meta(renderer_context):
    with pytest.deprecated_call():
        # pylint: disable-next=import-outside-toplevel,no-name-in-module
        from lektor.markdown import ImprovedRenderer
    improved_renderer = ImprovedRenderer()

    with pytest.deprecated_call(
        match=r"Use .*Renderer\.lektor\.meta instead"
    ) as warnings:
        assert improved_renderer.meta is renderer_context.meta
    assert all(w.filename == __file__ for w in warnings)


def test_deprecated_MarkdownConfig(improved_renderer):
    with pytest.deprecated_call(
        match=r"lektor\.markdown\.MarkdownConfig\b.* deprecated"
    ) as warnings:
        config = lektor.markdown.MarkdownConfig()
    assert all(w.filename == __file__ for w in warnings)
    assert config.renderer_base is type(improved_renderer)


def test_deprecated_escape(improved_renderer):
    with pytest.deprecated_call(
        match=r"lektor\.markdown\.escape\b.* deprecated"
    ) as warnings:
        assert lektor.markdown.escape("<") == "&lt;"
    assert all(w.filename == __file__ for w in warnings)


def test_getattr_raises():
    with pytest.raises(AttributeError, match="no attribute 'goober'"):
        _ = lektor.markdown.goober


def test_dir():
    expected = {
        "MISTUNE_VERSION",
        "Markdown",
        "get_controller",
        "escape",
        "ImprovedRenderer",
        "MarkdownConfig",
        "make_markdown",
        "markdown_to_html",
    }
    assert expected.issubset(dir(lektor.markdown))


def test_deprecated_make_markdown(env):
    with pytest.deprecated_call() as warnings:
        md = make_markdown(env)
    assert all(w.filename == __file__ for w in warnings)
    assert md("foo").strip() == "<p>foo</p>"


@pytest.mark.usefixtures("context")
def test_deprecated_markdown_to_html(record, field_options):
    with pytest.deprecated_call() as warnings:
        result = markdown_to_html("goober", record, field_options)
    assert all(w.filename == __file__ for w in warnings)
    assert result.html.rstrip() == "<p>goober</p>"


@pytest.mark.skipif(
    not MISTUNE_VERSION.startswith("0."),
    reason="Legacy renderer mixins will only work with mistune 0.x",
)
@pytest.mark.usefixtures("context")
def test_legacy_plugin(env, record, field_options):
    # Test that Lektor<3.4 style plugin works

    class LegacyPlugin(Plugin):
        """A legacy, Lektor<3.4, mistune 0.x style plugin."""

        name = "legacy-plugin"

        def on_markdown_config(self, config, **extra):
            # pylint: disable-next=import-outside-toplevel,no-name-in-module
            from lektor.markdown import escape

            class LegacyRendererMixin:
                def link(self, link, title, text):
                    self.meta["links"].append(link)
                    self.meta["record"] = self.record
                    return f'<a href="{escape(link)}">{text}</a>'

            config.renderer_mixins.append(LegacyRendererMixin)

        @staticmethod
        def on_markdown_meta_init(meta, **extra):
            meta["links"] = []

        @staticmethod
        def on_markdown_meta_postprocess(meta, **extra):
            meta["nlinks"] = len(meta["links"])

    env.plugin_controller.instanciate_plugin("legacy-plugin", LegacyPlugin)

    source = "[`foo`](https://example.org/)"
    markdown = Markdown(source, record, field_options)

    with pytest.deprecated_call():
        assert markdown["record"] is record
    assert markdown["links"] == ["https://example.org/"]
    assert markdown["nlinks"] == 1
    html = markdown.__html__().rstrip()
    assert html == '<p><a href="https://example.org/"><code>foo</code></a></p>'



import pytest


@pytest.mark.parametrize(
    "key1, key2, is_eq",
    [
        ({"path": "/"}, {"path": "/"}, True),
        ({"path": "/"}, {"path": "/", "alt": "en"}, True),
        ({"path": "/"}, {"path": "/blog"}, False),
        ({"path": "/blog/post1"}, {"path": "/blog/post1@siblings"}, False),
        ({"path": "/"}, {"path": "/", "page_num": 1}, False),
        ({"path": "/"}, {"path": "/", "alt": "de"}, False),
        ({"path": "/", "alt": "en"}, {"path": "/", "alt": "de"}, False),
    ],
)
def test_records_eq(pad, key1, key2, is_eq):
    r1 = pad.get(**key1)
    pad.cache.flush()
    r2 = pad.get(**key2)
    if is_eq:
        assert r1 == r2
        assert hash(r1) == hash(r2)
    else:
        assert r1 != r2


def test_records_from_different_pads_ne(env):
    pad1 = env.new_pad()
    pad2 = env.new_pad()
    assert pad1.get("/") == pad1.get("/")
    assert pad1.get("/") != pad2.get("/")


def test_asset_ne_record(pad):
    record = pad.get("/")
    asset = pad.get_asset("/")
    assert record != asset
    assert asset != record



from __future__ import annotations

from typing import Any
from typing import NamedTuple
from typing import TYPE_CHECKING

from flask import current_app
from flask import Flask
from flask import g
from werkzeug.utils import cached_property

from lektor.builder import Builder
from lektor.buildfailures import FailureController
from lektor.db import Database
from lektor.db import Pad
from lektor.db import Tree
from lektor.environment import Environment
from lektor.environment.config import Config
from lektor.reporter import CliReporter


if TYPE_CHECKING:
    from _typeshed import StrPath


class LektorInfo(NamedTuple):
    env: Environment
    output_path: StrPath
    verbosity: int = 0
    extra_flags: dict[str, str] | None = None


class LektorContext(LektorInfo):
    """Per-request object which provides the interface to Lektor for the Flask app(s).

    This does not provide any logic.  It just provides access to the
    needed Lektor internals and instances.
    """

    @property
    def project_id(self) -> str:
        return self.env.project.id

    @cached_property
    def database(self) -> Database:
        return Database(self.env)

    @cached_property
    def pad(self) -> Pad:
        return self.database.new_pad()

    @cached_property
    def tree(self) -> Tree:
        return Tree(self.pad)

    @property
    def config(self) -> Config:
        return self.database.config

    @cached_property
    def builder(self) -> Builder:
        return Builder(self.pad, self.output_path, extra_flags=self.extra_flags)

    @cached_property
    def failure_controller(self) -> FailureController:
        return FailureController(self.pad, self.output_path)

    def cli_reporter(self) -> CliReporter:
        return CliReporter(self.env, verbosity=self.verbosity)


class LektorApp(Flask):
    """A Flask app that has a lektor_info attribute."""

    def __init__(
        self,
        lektor_info: LektorInfo,
        **kwargs: Any,
    ) -> None:
        Flask.__init__(self, "lektor.admin", **kwargs)
        self.lektor_info = lektor_info


def get_lektor_context() -> LektorContext:
    if not hasattr(g, "lektor_context"):
        assert isinstance(current_app, LektorApp)
        lektor_info = current_app.lektor_info
        # pylint: disable=assigning-non-slot
        g.lektor_context = LektorContext._make(lektor_info)
    return g.lektor_context



from __future__ import annotations

import hashlib
import os
import sys
from enum import Enum
from pathlib import Path

from inifile import IniFile
from werkzeug.utils import cached_property

from lektor.environment import Environment
from lektor.utils import comma_delimited
from lektor.utils import get_cache_dir
from lektor.utils import untrusted_to_os_path


class Project:
    def __init__(self, name, project_file, tree, themes=None):
        self.name = name
        self.project_file = project_file
        self.tree = os.path.normpath(tree)
        self.themes = themes or []
        self.id = hashlib.md5(self.tree.encode("utf-8")).hexdigest()

    def open_config(self):
        if self.project_file is None:
            raise RuntimeError("This project has no project file.")
        return IniFile(self.project_file)

    @classmethod
    def from_file(cls, filename):
        """Reads a project from a project file."""
        inifile = IniFile(filename)
        if inifile.is_new:
            return None

        name = (
            inifile.get("project.name")
            or os.path.basename(filename).rsplit(".")[0].title()
        )
        path = os.path.join(
            os.path.dirname(filename),
            untrusted_to_os_path(inifile.get("project.path") or "."),
        )

        themes = inifile.get("project.themes")
        if themes is not None:
            themes = [x.strip() for x in themes.split(",")]
        else:
            themes = []

        return cls(
            name=name,
            project_file=filename,
            tree=path,
            themes=themes,
        )

    @classmethod
    def from_path(cls, path, extension_required=False):
        """Locates the project for a path."""
        path = os.path.abspath(path)
        if os.path.isfile(path) and (
            not extension_required or path.endswith(".lektorproject")
        ):
            return cls.from_file(path)

        try:
            files = [
                x for x in os.listdir(path) if x.lower().endswith(".lektorproject")
            ]
        except OSError:
            return None

        if len(files) == 1:
            return cls.from_file(os.path.join(path, files[0]))

        if os.path.isdir(path) and os.path.isfile(
            os.path.join(path, "content/contents.lr")
        ):
            return cls(
                name=os.path.basename(path),
                project_file=None,
                tree=path,
            )
        return None

    @classmethod
    def discover(cls, base=None):
        """Auto discovers the closest project."""
        if base is None:
            base = os.getcwd()
        here = base
        while 1:
            project = cls.from_path(here, extension_required=True)
            if project is not None:
                return project
            node = os.path.dirname(here)
            if node == here:
                break
            here = node
        return None

    @property
    def project_path(self):
        return self.project_file or self.tree

    def get_output_path(self):
        """The path where output files are stored."""
        config = self.open_config()  # raises if no project_file
        output_path = config.get("project.output_path")
        if output_path:
            path = Path(config.filename).parent / output_path
        else:
            path = Path(get_cache_dir(), "builds", self.id)
        return str(path)

    class PackageCacheType(Enum):
        VENV = "venv"  # The new virtual environment-based package cache
        FLAT = "flat"  # No longer used flat-directory package cache

    def get_package_cache_path(
        self, cache_type: PackageCacheType = PackageCacheType.VENV
    ) -> Path:
        """The path where plugin packages are stored."""
        if cache_type is self.PackageCacheType.FLAT:
            cache_name = "packages"
        else:
            cache_name = "venvs"

        h = hashlib.md5()
        h.update(self.id.encode("utf-8"))
        h.update(sys.version.encode("utf-8"))
        h.update(sys.prefix.encode("utf-8"))

        return Path(get_cache_dir(), cache_name, h.hexdigest())

    def content_path_from_filename(self, filename):
        """Given a filename returns the content path or None if
        not in project.
        """
        dirname, basename = os.path.split(os.path.abspath(filename))
        if basename == "contents.lr":
            path = dirname
        elif basename.endswith(".lr"):
            path = os.path.join(dirname, basename[:-3])
        else:
            return None

        content_path = os.path.normpath(self.tree).split(os.path.sep) + ["content"]
        file_path = os.path.normpath(path).split(os.path.sep)
        prefix = os.path.commonprefix([content_path, file_path])
        if prefix == content_path:
            return "/" + "/".join(file_path[len(content_path) :])
        return None

    def make_env(self, load_plugins=True):
        """Create a new environment for this project."""
        return Environment(self, load_plugins=load_plugins)

    @cached_property
    def excluded_assets(self):
        """List of glob patterns matching filenames of excluded assets.

        Combines with default EXCLUDED_ASSETS.
        """
        config = self.open_config()
        return list(comma_delimited(config.get("project.excluded_assets", "")))

    @cached_property
    def included_assets(self):
        """List of glob patterns matching filenames of included assets.

        Overrides both excluded_assets and the default excluded patterns.
        """
        config = self.open_config()
        return list(comma_delimited(config.get("project.included_assets", "")))

    def to_json(self):
        return {
            "name": self.name,
            "project_file": self.project_file,
            "project_path": self.project_path,
            "default_output_path": self.get_output_path(),
            "package_cache_path": str(self.get_package_cache_path()),
            "id": self.id,
            "tree": self.tree,
        }



import os
import shutil
import textwrap

import pytest

from lektor.assets import Asset
from lektor.builder import Builder
from lektor.db import Database
from lektor.environment import Environment
from lektor.project import Project
from lektor.reporter import Reporter


sep = os.path.sep


@pytest.fixture(scope="function")
def theme_project_tmpdir(tmp_path, data_path):
    """Copy themes-project to a temp dir, and copy demo-project content to it."""
    temp_dir = tmp_path / "temp/themes-project"
    temp_dir.parent.mkdir()
    shutil.copytree(data_path / "themes-project", temp_dir)
    shutil.copytree(data_path / "demo-project/content", temp_dir / "content")
    return temp_dir


@pytest.fixture(scope="function")
def theme_project(theme_project_tmpdir, request):
    """Return the theme project created in a temp dir.

    Could be parametrize, if request.param=False themes variables won't be set
    """
    try:
        with_themes_var = request.param
    except AttributeError:
        with_themes_var = True

    # Create the .lektorproject file
    lektorfile_text = textwrap.dedent(
        """
        [project]
        name = Themes Project
        {}
    """.format("themes = blog_theme, project_theme" if with_themes_var else "")
    )
    lektorfile = theme_project_tmpdir / "themes.lektorproject"
    lektorfile.write_text(lektorfile_text, "utf8")
    return Project.from_path(theme_project_tmpdir)


@pytest.fixture(scope="function")
def theme_env(theme_project, save_sys_path):
    return Environment(theme_project)


@pytest.fixture(scope="function")
def theme_pad(theme_env):
    return Database(theme_env).new_pad()


@pytest.fixture(scope="function")
def theme_builder(theme_pad, tmpdir):
    return Builder(theme_pad, str(tmpdir.mkdir("output")))


@pytest.mark.parametrize(
    "theme_project, themes",
    [
        (True, ["blog_theme", "project_theme"]),
        # when themes variables isn't set themes are only loaded in the env
        (False, []),
    ],
    indirect=["theme_project"],
)
def test_loading_theme_variable(theme_project, themes):
    assert theme_project.themes == themes


@pytest.mark.parametrize("theme_project", (True,), indirect=True)
def test_loading_theme_path(theme_env):
    assert [os.path.basename(p) for p in theme_env.theme_paths] == [
        "blog_theme",
        "project_theme",
    ]


@pytest.mark.parametrize("theme_project", (False,), indirect=True)
def test_loading_theme_path_variable_dont_set(theme_env):
    """Themes will be loaded, but the order could change."""
    paths = [os.path.basename(p) for p in theme_env.theme_paths]
    assert len(paths) == 2
    for path in paths:
        assert path in ["blog_theme", "project_theme"]


@pytest.mark.parametrize(
    "asset_name, found_in",
    [
        # - themes-project/assets/asset.txt
        # only exist in themes-project assets will be loaded from there
        ("asset.txt", "root"),
        # - themes-project/assets/dummy.txt
        # - themes-project/themes/blog_theme/assets/dummy.txt
        # wil be loaded from themes-project assets not from blog_theme assets
        ("dummy.txt", "root"),
        # - themes-project/themes/blog_theme/static/blog.css
        # only exist in blog_theme assets will be loaded from there
        ("static/blog.css", "blog"),
        # - themes-project/themes/project_theme/static/project.css
        # only exist in project_theme assets will be loaded from there
        ("static/project.css", "project"),
        # - themes-project/themes/blog_theme/assets/dummy2.txt
        # - themes-project/themes/project_theme/assets/dummy2.txt
        # wil be loaded from blog_theme assets because is included first
        ("dummy2.txt", "blog"),
    ],
)
def test_theme_asset_loading(theme_pad, asset_name, found_in):
    """Test loading assets from theme project.

    Loading should take in account the order of the themes
    """
    path_list = theme_pad.get_asset(asset_name).source_filename.split(sep)
    path_list_from_url = theme_pad.resolve_url_path(asset_name).source_filename.split(
        sep
    )

    assert path_list == path_list_from_url

    assert (found_in == "root") == ("themes" not in path_list)
    assert (found_in == "blog") == ("blog_theme" in path_list)
    assert (found_in == "project") == ("project_theme" in path_list)


@pytest.mark.parametrize(
    "url, datamodel_name, found_in",
    [
        # - themes-project/themes/blog_theme/models/blog.ini
        # only exist in blog_theme will be loaded from there
        ("/blog", "Blog", "blog"),
        # - themes-project/themes/project_theme/models/projects.ini
        # only exist in project_theme will be loaded from there
        ("/projects", "Projects", "project"),
        # - themes-project/models/blog-post.ini
        # - themes-project/themes/blog_theme/models/blog-post.ini
        # will be loaded from themes-project models
        ("/blog/post1", "Blog Post", "root"),
        # - themes-project/models/page.ini
        # only exist in themes-project will be loaded from there
        ("/", "Page", "root"),
        # - themes-project/themes/blog_theme/models/project.ini
        # - themes-project/themes/project_theme/models/project.ini
        # wil be loaded from blog_theme assets because is included first
        ("/projects/bagpipe", "Project", "blog"),
    ],
)
def test_theme_models_loading(theme_pad, url, datamodel_name, found_in):
    """Test loading models from theme project.

    Loading should take in account the order of the themes
    """
    assert datamodel_name == theme_pad.get(url).datamodel.name

    path_list = theme_pad.get(url).datamodel.filename.split(sep)

    assert (found_in == "root") == ("themes" not in path_list)
    assert (found_in == "blog") == ("blog_theme" in path_list)
    assert (found_in == "project") == ("project_theme" in path_list)


@pytest.mark.parametrize(
    "template_name, found_in",
    [
        # - themes-project/templates/layout.html
        # - themes-project/themes/blog_theme/templates/layout.html
        # - themes-project/themes/project_theme/templates/layout.html
        # will be loaded from themes-project templates
        ("layout.html", "root"),
        # - themes-project/themes/blog_theme/templates/blog.html
        # only exist in blog_theme will be loaded from there
        ("blog.html", "blog"),
        # - themes-project/themes/project_theme/templates/project.html
        # only exist in project_theme will be loaded from there
        ("project.html", "project"),
        # - themes-project/templates/page.html
        # only exist in themes-project will be loaded from there
        ("page.html", "root"),
        # - themes-project/themes/blog_theme/templates/dummy.html
        # - themes-project/themes/project_theme/templates/dummy.html
        # wil be loaded from blog_theme assets because is included first
        ("dummy.html", "blog"),
    ],
)
def test_theme_templates_loading(theme_env, template_name, found_in):
    """Test loading templates from theme project.

    Loading should take in account the order of the themes
    """
    path_list = theme_env.jinja_env.get_template(template_name).filename.split(sep)

    assert (found_in == "root") == ("themes" not in path_list)
    assert (found_in == "blog") == ("blog_theme" in path_list)
    assert (found_in == "project") == ("project_theme" in path_list)


class BuiltAssetsReporter(Reporter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.built_assets = []

    def start_artifact_build(self, is_current):
        source_obj = self.current_artifact.source_obj
        if isinstance(source_obj, Asset):
            self.built_assets.append(source_obj)


def test_build_omits_shadowed_assets(theme_builder, theme_env):
    with BuiltAssetsReporter(theme_env) as reporter:
        theme_builder.build_all()
    asset_urls = [asset.url_path for asset in reporter.built_assets]
    assert len(asset_urls) > 0
    assert len(set(asset_urls)) == len(asset_urls)



import flask
import pytest

import lektor.admin.webui


@pytest.fixture
def output_path(tmp_path):
    output_path = tmp_path / "output_path"
    output_path.mkdir()
    return output_path


@pytest.fixture
def static_folder(tmp_path) -> str:
    scratch_folder = tmp_path / "static"
    scratch_folder.mkdir()
    return scratch_folder


@pytest.fixture
def test_client(env, static_folder, output_path):
    app = lektor.admin.webui.make_app(
        env, output_path=output_path, static_folder=static_folder
    )
    with app.test_client() as client:
        yield client


@pytest.fixture
def app_static_dummy_txt(static_folder):
    """Create a dummy file in the Flask apps static directory."""
    dummy_txt = static_folder / "dummy.txt"
    dummy_txt.touch()


################################################################


@pytest.mark.parametrize(
    "url, mimetype, is_editable",
    [
        ("/projects/coffee/", "text/html", True),  # Page
        ("/hello.txt", "text/plain", False),  # Attachment
        # Asset file
        ("/static/demo.css", "text/css", False),
        # Asset directories with index.{htm,html}
        ("/dir_with_index_html/", "text/html", False),
        ("/dir_with_index_htm/", "text/html", False),
        ("/dir_with_index_html/index.html", "text/html", False),
        ("/dir_with_index_htm/index.htm", "text/html", False),
        # Flask static directory
        ("/admin/static/dummy.txt", "text/plain", False),
    ],
)
@pytest.mark.usefixtures("app_static_dummy_txt")
def test_get(test_client, url, mimetype, is_editable):
    resp = test_client.get(url)
    assert resp.status_code == 200
    assert resp.mimetype == mimetype
    data = b"".join(resp.get_app_iter(flask.request.environ)).decode("utf-8")
    assert ("/admin/edit?" in data) == is_editable


def test_get_admin_does_something_useful(test_client, mocker):
    # Test that GET /admin eventually gets to the admin JS app
    # See https://github.com/lektor/lektor/issues/1043
    render_template = mocker.patch(
        "lektor.admin.modules.dash.render_template",
        return_value="RENDERED",
    )
    resp = test_client.get("/admin", follow_redirects=True)
    assert resp.status_code == 200
    assert resp.get_data(as_text=True) == render_template.return_value
    assert render_template.mock_calls == [
        mocker.call("dash.html", lektor_config=mocker.ANY),
    ]



from jinja2 import Undefined

from lektor.constants import PRIMARY_ALT


class BadValue(Undefined):
    __slots__ = ()


def get_undefined_info(undefined):
    if isinstance(undefined, Undefined):
        try:
            undefined._fail_with_undefined_error()
        except Exception as e:
            return str(e)
    return "defined value"


class RawValue:
    __slots__ = ("name", "value", "field", "pad")

    def __init__(self, name, value=None, field=None, pad=None):
        self.name = name
        self.value = value
        self.field = field
        self.pad = pad

    def _get_hint(self, prefix, reason):
        if self.field is not None:
            return f"{prefix} in field '{self.field.name}': {reason}"
        return f"{prefix}: {reason}"

    def bad_value(self, reason):
        return BadValue(hint=self._get_hint("Bad value", reason), obj=self.value)

    def missing_value(self, reason):
        return Undefined(hint=self._get_hint("Missing value", reason), obj=self.value)


class _NameDescriptor:
    def __get__(self, obj, type):
        rv = type.__name__
        rv = rv.removesuffix("Type")
        return rv.lower()


class Type:
    widget = "multiline-text"

    def __init__(self, env, options):
        self.env = env
        self.options = options

    @property
    def size(self):
        size = self.options.get("size") or "normal"
        if size not in ("normal", "small", "large"):
            size = "normal"
        return size

    @property
    def width(self):
        return self.options.get("width") or "1/1"

    name = _NameDescriptor()

    def to_json(self, pad, record=None, alt=PRIMARY_ALT):
        return {
            "name": self.name,
            "widget": self.widget,
            "size": self.size,
            "width": self.width,
        }

    def value_from_raw(self, raw):
        # pylint: disable=no-self-use
        return raw

    def value_from_raw_with_default(self, raw):
        value = self.value_from_raw(raw)
        if (
            isinstance(value, Undefined)
            and raw.field is not None
            and raw.field.default is not None
        ):
            return self.value_from_raw(
                RawValue(raw.name, raw.field.default, field=raw.field, pad=raw.pad)
            )
        return value

    def __repr__(self):
        return f"{self.__class__.__name__}()"



import pytest

from lektor.editor import BadEdit
from lektor.editor import make_editor_session
from lektor.editor import MutableEditorData


@pytest.mark.parametrize(
    "path, kwargs, expect",
    [
        ("new", {}, {"exists": False, "datamodel": "page"}),
        ("new", {"alt": "en"}, {"exists": False, "datamodel": "page"}),
        ("projects/new", {}, {"exists": False, "datamodel": "project"}),
        ("projects/new", {"datamodel": "page"}, {"exists": False, "datamodel": "page"}),
        ("projects/zaun", {"alt": "de"}, {"exists": True, "datamodel": "project"}),
        ("projects/zaun", {"alt": "en"}, {"exists": False}),
        ("projects/zaun", {}, {"exists": False}),
        ("projects/zaun", {"alt": "_primary"}, {"exists": False}),
    ],
)
def test_make_editor_session(pad, path, kwargs, expect):
    sess = make_editor_session(pad, path, **kwargs)
    if "exists" in expect:
        assert sess.exists == expect["exists"]
    if "datamodel" in expect:
        assert sess.datamodel.id == expect["datamodel"]


@pytest.mark.parametrize(
    "path, kwargs, expect",
    [
        ("projects/zaun", {"alt": "xx"}, "invalid alternative"),
        ("projects/.zaun", {}, "Invalid ID"),
        ("projects", {"is_attachment": True}, "attachment flag"),
        ("projects", {"datamodel": "page"}, "datamodel"),
        pytest.param(
            # model conflict with that of existing alt
            #
            # Different alts of the same page should not be able to have different
            # models, right?
            "projects/zaun",
            {"alt": "en", "datamodel": "page"},
            "conflicting",
            marks=pytest.mark.xfail(reason="buglet that should be fixed"),
        ),
        ("/../../templates", {}, "Invalid path"),
    ],
)
def test_make_editor_session_raises_bad_edit(pad, path, kwargs, expect):
    with pytest.raises(BadEdit) as excinfo:
        make_editor_session(pad, path, **kwargs)
    assert expect in str(excinfo.value)


def test_basic_editor(scratch_tree):
    sess = scratch_tree.edit("/")

    assert sess.id == ""
    assert sess.path == "/"
    assert sess.record is not None

    assert sess.data["_model"] == "page"
    assert sess.data["title"] == "Index"
    assert sess.data["body"] == "*Hello World!*"

    sess.data["body"] = "A new body"
    sess.commit()

    assert sess.closed

    with open(sess.get_fs_path(), encoding="utf-8") as f:
        assert f.read().splitlines() == [
            "_model: page",
            "---",
            "title: Index",
            "---",
            "body: A new body",
        ]


def test_create_alt(scratch_tree, scratch_pad):
    sess = scratch_tree.edit("/", alt="de")

    assert sess.id == ""
    assert sess.path == "/"
    assert sess.record is not None

    assert sess.data["_model"] == "page"
    assert sess.data["title"] == "Index"
    assert sess.data["body"] == "*Hello World!*"

    sess.data["body"] = "Hallo Welt!"
    sess.commit()

    assert sess.closed

    # When we use the editor to change this, we only want the fields that
    # changed compared to the base to be included.
    with open(sess.get_fs_path(alt="de"), encoding="utf-8") as f:
        assert f.read().splitlines() == ["body: Hallo Welt!"]

    scratch_pad.cache.flush()
    item = scratch_pad.get("/", alt="de")
    assert item["_slug"] == ""
    assert item["title"] == "Index"
    assert item["body"].source == "Hallo Welt!"
    assert item["_model"] == "page"


@pytest.fixture
def original_data():
    return {"_id": "test", "item1": "value1"}


@pytest.fixture
def fallback_data():
    return {"_id": "test", "item1": "fallback1", "item2": "fallback2"}


@pytest.fixture
def editor_data(original_data, fallback_data):
    return MutableEditorData(original_data, fallback_data)


def test_editor_data_ischanged(editor_data):
    assert not editor_data.ischanged()
    editor_data["item2"] = "fallback2"
    assert not editor_data.ischanged()
    editor_data["item2"] = "new2"
    assert editor_data.ischanged()
    editor_data["item2"] = "fallback2"
    assert not editor_data.ischanged()


def test_editor_data_getitem(editor_data):
    assert editor_data["_id"] == "test"
    assert editor_data["item1"] == "value1"
    assert editor_data["item2"] == "fallback2"
    with pytest.raises(KeyError):
        _ = editor_data["item3"]


def test_editor_data_setitem(editor_data):
    with pytest.raises(KeyError):
        editor_data["_path"] = "newpath"
    assert not editor_data.ischanged()

    editor_data["item1"] = "newval"
    assert editor_data["item1"] == "newval"
    assert editor_data.ischanged()

    editor_data["item1"] = "value1"
    assert editor_data["item1"] == "value1"
    assert not editor_data.ischanged()


def test_editor_data_setitem_possibly_implied_key(editor_data, original_data):
    original_data["_model"] = "mymodel"
    assert not editor_data.ischanged()
    editor_data["_model"] = "mymodel"
    assert editor_data.ischanged()


def test_editor_data_delitem(editor_data):
    del editor_data["item1"]
    assert editor_data.ischanged()
    assert "item1" not in editor_data

    editor_data["item1"] = "newval"
    assert editor_data.ischanged()
    assert "item1" in editor_data

    editor_data["item1"] = "value1"
    assert not editor_data.ischanged()
    assert "item1" in editor_data


def test_editor_data_delitem_raises_keyerror(editor_data):
    with pytest.raises(KeyError):
        del editor_data["missing"]
    with pytest.raises(KeyError):
        del editor_data["_id"]


def test_revert_key(editor_data):
    editor_data["item2"] = "xx"
    assert editor_data.ischanged()
    editor_data.revert_key("item2")
    assert not editor_data.ischanged()
    assert editor_data["item2"] == "fallback2"


def test_editor_data_iter(editor_data):
    assert list(editor_data) == ["item1", "item2"]
    editor_data["item3"] = "new3"
    assert list(editor_data) == ["item1", "item2", "item3"]
    del editor_data["item1"]
    assert list(editor_data) == ["item2", "item3"]


def test_editor_data_len(editor_data):
    assert len(editor_data) == 2
    del editor_data["item1"]
    assert len(editor_data) == 1
    editor_data["item3"] = "new3"
    assert len(editor_data) == 2


def test_editor_data_keys(editor_data):
    assert list(editor_data.keys()) == ["item1", "item2"]
    assert "item1" in editor_data.keys()
    assert len(editor_data.keys()) == 2

    assert list(editor_data.keys(fallback=False)) == ["item1"]
    assert "item2" not in editor_data.keys(fallback=False)
    assert len(editor_data.keys(fallback=False)) == 1


def test_editor_data_items(editor_data):
    assert list(editor_data.items(fallback=False)) == [
        ("item1", "value1"),
    ]
    assert ("item2", "fallback2") not in editor_data.items(fallback=False)
    assert len(editor_data.items(fallback=False)) == 1

    editor_data["item3"] = "x3"
    assert list(editor_data.items()) == [
        ("item1", "value1"),
        ("item2", "fallback2"),
        ("item3", "x3"),
    ]
    assert ("item2", "fallback2") in editor_data.items()
    assert ("item3", "x3") in editor_data.items()
    assert len(editor_data.items()) == 3


def test_editor_data_values(editor_data):
    del editor_data["item1"]
    assert list(editor_data.values()) == ["fallback2"]
    assert "fallback2" in editor_data.values()
    assert len(editor_data.values()) == 1

    assert not list(editor_data.values(fallback=False))
    assert "fallback2" not in editor_data.values(fallback=False)
    assert len(editor_data.values(fallback=False)) == 0


def test_deprecated_data_proxy_methods(pad):
    editor = make_editor_session(pad, "/")

    with pytest.deprecated_call(match=r"EditorSession\.__contains__ .* deprecated"):
        assert "arbitrary" not in editor

    with pytest.deprecated_call(match=r"EditorSession\.__setitem__ .* deprecated"):
        editor["test"] = "value"
    assert editor.data["test"] == "value"
    with pytest.deprecated_call(match=r"EditorSession\.__getitem__ .* deprecated"):
        assert editor["test"] == "value"

    with pytest.deprecated_call(match=r"EditorSession\.update .* deprecated"):
        editor.update({"test": "another"})
    assert editor.data["test"] == "another"

    items = set(editor.data.items())
    with pytest.deprecated_call(match=r"EditorSession\.__len__ .* deprecated"):
        assert len(editor) == len(items)
    with pytest.deprecated_call(match=r"EditorSession\.items .* deprecated"):
        assert set(editor.items()) == items
    with pytest.deprecated_call(match=r"EditorSession\.__iter__ .* deprecated"):
        assert set(editor) == {key for key, val in items}
    with pytest.deprecated_call(match=r"EditorSession\.keys .* deprecated"):
        assert set(editor.keys()) == {key for key, val in items}
    with pytest.deprecated_call(match=r"EditorSession\.values .* deprecated"):
        assert set(editor.values()) == {val for key, val in items}

    with pytest.deprecated_call(match=r"EditorSession\.iteritems .*\.data\.items "):
        assert set(editor.iteritems()) == items
    with pytest.deprecated_call(match=r"EditorSession\.iterkeys .*\.data\.keys "):
        assert set(editor.iterkeys()) == {key for key, val in items}
    with pytest.deprecated_call(match=r"EditorSession\.itervalues .*\.data\.values "):
        assert set(editor.itervalues()) == {val for key, val in items}

    with pytest.deprecated_call(match=r"EditorSession\.revert_key .* deprecated"):
        editor.revert_key("test")
    assert "test" not in editor.data



from pathlib import Path

import pytest

from lektor.builder import Builder
from lektor.builder import FileInfo
from lektor.project import Project
from lektor.reporter import NullReporter


def get_child_sources(prog):
    return sorted(prog.iter_child_sources(), key=lambda x: x["_id"])


def test_basic_build(pad, builder):
    root = pad.root

    prog, build_state = builder.build(root)
    assert prog.source is root
    assert build_state.failed_artifacts == []

    (artifact,) = prog.artifacts
    # Root and its thumbnail image were updated.
    assert artifact in build_state.updated_artifacts
    assert artifact.artifact_name == "index.html"
    assert set(artifact.sources) == set(root.iter_source_filenames())
    assert artifact.updated
    assert artifact.extra is None
    assert artifact.config_hash is None


def test_child_sources_basic(pad, builder):
    extra = pad.get("/extra")

    prog, _ = builder.build(extra)
    child_sources = get_child_sources(prog)

    assert [x["_id"] for x in child_sources] == [
        "a",
        "b",
        "container",  # hidden children should be built, too
        "file.ext",
        "hello.txt",
        "paginated",
        "slash-slug",
    ]


def test_child_sources_pagination(pad, builder):
    projects = pad.get("/projects")

    def ids(sources):
        return [x["_id"] for x in sources]

    prog, _ = builder.build(projects)

    child_sources = get_child_sources(prog)
    assert ids(child_sources) == [
        "attachment.txt",
        "filtered",
        "projects",
        "projects",
        "secret",
    ]

    page1 = child_sources[2]
    assert page1["_id"] == "projects"
    assert page1.page_num == 1
    page2 = child_sources[3]
    assert page2["_id"] == "projects"
    assert page2.page_num == 2

    prog, _ = builder.build(page1)
    child_sources_p1 = get_child_sources(prog)

    assert ids(child_sources_p1) == [
        "bagpipe",
        "coffee",
        "master",
        "oven",
    ]

    prog, _ = builder.build(page2)
    child_sources_p2 = get_child_sources(prog)

    assert ids(child_sources_p2) == [
        "postage",
        "slave",
        "wolf",
    ]


def test_basic_artifact_current_test(pad, builder, reporter):
    post1 = pad.get("blog/post1")

    def build():
        reporter.clear()
        prog, _ = builder.build(post1)
        return prog.artifacts[0]

    artifact = build()

    assert reporter.get_major_events() == [
        (
            "enter-source",
            {
                "source": post1,
            },
        ),
        (
            "start-artifact-build",
            {
                "artifact": artifact,
                "is_current": False,
            },
        ),
        (
            "build-func",
            {
                "func": "lektor.build_programs.PageBuildProgram",
            },
        ),
        (
            "finish-artifact-build",
            {
                "artifact": artifact,
            },
        ),
        (
            "leave-source",
            {
                "source": post1,
            },
        ),
    ]

    assert set(reporter.get_recorded_dependencies()) == {
        "Website.lektorproject",
        "content/blog/post1/contents+en.lr",
        "content/blog/post1/contents.lr",
        "content/blog/post1/hello.txt+en.lr",
        "content/blog/post1/hello.txt.lr",
        "templates/blog-post.html",
        "templates/layout.html",
        "models/blog-post.ini",
    }

    assert artifact.is_current

    artifact = build()

    assert artifact.is_current

    assert reporter.get_major_events() == [
        (
            "enter-source",
            {
                "source": post1,
            },
        ),
        (
            "start-artifact-build",
            {
                "artifact": artifact,
                "is_current": True,
            },
        ),
        (
            "build-func",
            {
                "func": "lektor.build_programs.PageBuildProgram",
            },
        ),
        (
            "finish-artifact-build",
            {
                "artifact": artifact,
            },
        ),
        (
            "leave-source",
            {
                "source": post1,
            },
        ),
    ]


def test_basic_template_rendering(pad, builder):
    root = pad.root

    prog, _ = builder.build(root)
    artifact = prog.artifacts[0]

    with artifact.open("rb") as f:
        rv = f.read().decode("utf-8")

    assert artifact.artifact_name == "index.html"

    assert "<title>My Website</title>" in rv
    assert "<h1>Welcome</h1>" in rv
    assert '<link href="static/style.css" rel="stylesheet">' in rv
    assert "<p>Welcome to this pretty nifty website.</p>" in rv


def test_attachment_copying(pad, builder):
    root = pad.root
    text_file = root.attachments.get("hello.txt")

    prog, _ = builder.build(text_file)
    artifact = prog.artifacts[0]

    assert artifact.artifact_name == "hello.txt"

    with artifact.open("rb") as f:
        rv = f.read().decode("utf-8").strip()
        assert rv == "Hello I am an Attachment"


def test_asset_processing(pad, builder):
    static = pad.asset_root.get_child("static")

    prog, _ = builder.build(static)
    assets = list(prog.iter_child_sources())
    assert len(assets) == 1
    assert assets[0].name == "demo.css"

    prog, _ = builder.build(assets[0])
    with prog.artifacts[0].open("rb") as f:
        rv = f.read().decode("utf-8").strip()
        assert "color: red" in rv


def test_included_assets(pad, builder):
    # In demo-project/Website.lektorproject, included_assets = "_*".
    root = pad.asset_root

    prog, _ = builder.build(root)
    assets = list(prog.iter_child_sources())
    assert "_include_me_despite_underscore" in [a.name for a in assets]


def test_excluded_assets(pad, builder):
    # In demo-project/Website.lektorproject, excluded_assets = "foo*".
    root = pad.asset_root

    prog, _ = builder.build(root)
    assets = list(prog.iter_child_sources())
    assert "foo-prefix-makes-me-excluded" not in [a.name for a in assets]


def test_iter_child_pages(child_sources_test_project_builder):
    # Test that child sources are built even if they're filtered out by a
    # pagination query like "this.children.filter(F._model == 'doesnt-exist')".
    builder = child_sources_test_project_builder
    pad = builder.pad
    prog, _ = builder.build(pad.root)
    assert builder.pad.get("filtered-page") in prog.iter_child_sources()


def test_iter_child_attachments(child_sources_test_project_builder):
    # Test that attachments are built, even if a pagination has no items.
    builder = child_sources_test_project_builder
    pad = builder.pad
    prog, _ = builder.build(pad.root)
    assert builder.pad.get("attachment.txt") in prog.iter_child_sources()


@pytest.mark.parametrize(
    "parent_path, child_name",
    [
        ("/extra", "container"),  # a hidden page
        ("/extra/container", "a"),  # a page whose parent is hidden
        ("/extra/container", "hello.txt"),  # an attachment whose parent is hidden
    ],
)
def test_iter_children_of_hidden_pages(builder, pad, parent_path, child_name):
    # Test that child sources are built even if they're parent is hidden
    parent = pad.get(parent_path)
    child = pad.get(f"{parent_path}/{child_name}")
    # sanity checks
    assert parent is not None and child is not None
    assert parent.is_hidden or child.is_hidden

    prog, _ = builder.build(parent)
    assert child in prog.iter_child_sources()


def test_record_is_file(pad, builder):
    record = pad.get("/extra/file.ext")

    prog, _ = builder.build(record)
    (artifact,) = prog.artifacts
    assert artifact.artifact_name == "extra/file.ext"


def test_slug_contains_slash(pad, builder):
    record = pad.get("/extra/slash-slug")

    prog, _ = builder.build(record)
    (artifact,) = prog.artifacts
    assert artifact.artifact_name == "extra/long/path/index.html"


def test_asseturl_dependency_tracking_integration(
    scratch_project_data, scratch_pad, scratch_builder
):
    scratch_project_data.joinpath("templates/page.html").write_text(
        "{{ '/asset.txt'|asseturl }}", "utf-8"
    )
    asset_txt = scratch_project_data / "assets/asset.txt"
    asset_txt.parent.mkdir(exist_ok=True)
    asset_txt.write_text("an asset\n", "utf-8")

    prog, build_state = scratch_builder.build(scratch_pad.root)
    assert len(build_state.updated_artifacts) == 1
    output = Path(prog.artifacts[0].dst_filename)
    asset_url = output.read_text(encoding="utf-8").rstrip()

    prog, build_state = scratch_builder.build(scratch_pad.root)
    assert len(build_state.updated_artifacts) == 0

    asset_txt.write_text("updated asset\n", "utf-8")
    prog, build_state = scratch_builder.build(scratch_pad.root)
    updated_asset_url = output.read_text(encoding="utf-8").rstrip()
    assert updated_asset_url != asset_url
    assert len(build_state.updated_artifacts) == 1


def test_prune_remove_artifacts_of_hidden_pages(scratch_project_data, scratch_builder):
    pad = scratch_builder.pad
    # Build root page
    prog, _ = scratch_builder.build(pad.root)
    (artifact,) = prog.artifacts
    assert Path(artifact.dst_filename).is_file()

    # Do a prune.  Output artifact should survive
    pad.cache.flush()
    scratch_builder.prune()
    assert Path(artifact.dst_filename).is_file()

    # Mark page as hidden, prune should then clean the artifact
    contents_lr = scratch_project_data.joinpath("content/contents.lr")
    contents_lr.write_text(contents_lr.read_text() + "\n---\n_hidden: yes\n")
    pad.cache.flush()
    scratch_builder.prune()
    assert not Path(artifact.dst_filename).is_file()


def test_prune_all(builder):
    pad = builder.pad
    # Don't build / — it has a zillion thumbnail images and so is slow to build
    src = pad.get("/blog/post1")
    prog, _ = builder.build(src)
    (artifact,) = prog.artifacts
    assert Path(artifact.dst_filename).is_file()
    pad.cache.flush()
    builder.prune(all=True)
    assert not Path(artifact.dst_filename).is_file()


class AssertBuildsNothingReporter(NullReporter):
    """Reporter to collect source objects which are built during a build cycle."""

    def __init__(self):
        super().__init__(env=None)

    def start_artifact_build(self, is_current):
        assert is_current


def test_second_build_all_builds_nothing(scratch_builder, scratch_project_data):
    # This excercises a bug having to do with not properly tracking the alt
    # of virtual source object dependencies.  See #1108 (and #1007, #959).

    # The project should have multiple alts configured. (Scratch_project does.)

    # Add child pages to scratch_project
    for child in "child1", "child2":
        child_lr = scratch_project_data / "content" / child / "contents.lr"
        child_lr.parent.mkdir()
        child_lr.write_text("_template: child.html\n")
    # Reference siblings
    scratch_project_data.joinpath("templates/child.html").write_text(
        "{{ this.get_siblings() }}"
    )

    scratch_builder.build_all()

    with AssertBuildsNothingReporter():
        scratch_builder.build_all()


@pytest.mark.parametrize(
    "param",
    [
        "project",
        pytest.param(
            "symlinked-project",
            marks=pytest.mark.xfail(
                reason="FIXME: PathCache.to_source_filename does not cope with symlinks"
            ),
        ),
    ],
)
def test_BuildState_to_source_filename(param, data_path, tmp_path):
    demo_project = data_path / "demo-project"
    if "symlink" in param:
        tree = tmp_path / "tree"
        try:
            tree.symlink_to(demo_project, target_is_directory=True)
        except OSError:
            pytest.skip("symlinks unsupported?")
    else:
        tree = demo_project
    env = Project.from_path(tree).make_env(load_plugins=False)
    build_state = Builder(env.new_pad(), tmp_path / "output").new_build_state()

    assert build_state.to_source_filename(str(demo_project / "filename")) == "filename"


################################################################


def test_Artifact_open_encoding(builder):
    build_state = builder.new_build_state()
    artifact = build_state.new_artifact("dummy-artifact", sources=())
    with artifact.open("w", encoding="iso-8859-1") as fp:
        fp.write("Ciarán")
    with artifact.open("r", encoding="iso-8859-1") as fp:
        assert fp.read() == "Ciarán"


def test_Artifact_file_mode(builder, tmp_path):
    # Check that created artifacts have same file mode as normally created files.
    new_file = tmp_path / "dummy-test-file"
    new_file.touch()
    new_file_mode = new_file.stat().st_mode

    build_state = builder.new_build_state()
    artifact = build_state.new_artifact("dummy-artifact", sources=())
    with artifact.update(), artifact.open("w"):
        pass
    artifact_mode = Path(artifact.dst_filename).stat().st_mode

    # applying oct makes failures more readable
    assert oct(artifact_mode) == oct(new_file_mode)


def test_FileInfo_unchanged(env, tmp_path):
    file_path = tmp_path / "file"
    file_path.write_text("foo")

    file_info = FileInfo(env, file_path)
    # cache size, mtime, but *not* checksum
    assert file_info.size == 3

    file_path.write_text("foobar")
    file_info2 = FileInfo(env, file_path)
    assert file_info2.size != 3

    assert not file_info.unchanged(file_info2)



import importlib
import os
import shutil
import sys
import textwrap
from contextlib import contextmanager
from contextlib import suppress
from pathlib import Path

import pytest
from _pytest.monkeypatch import MonkeyPatch

import lektor.project
from lektor.builder import Builder
from lektor.db import Database
from lektor.db import Tree
from lektor.environment import Environment
from lektor.environment.expressions import Expression
from lektor.project import Project
from lektor.reporter import BufferReporter
from lektor.utils import locate_executable


@pytest.fixture(scope="session")
def data_path():
    """Path to directory which contains test data.

    Current this data lives in the ``tests`` directory.
    """
    return Path(__file__).parent


@pytest.fixture(scope="session", autouse=True)
def temporary_lektor_cache(tmp_path_factory):
    """Get Lektor to use a temporary cache directory.

    This prevents the tests from leaving scats behind in the
    user’s real cache directory.

    """
    cache_dir = tmp_path_factory.mktemp("lektor_cache")

    # The stock monkeypatch fixture is function-scoped and so can not
    # be used in a session-scoped fixture.
    # Workaround from:
    # https://github.com/pytest-dev/pytest/issues/363#issuecomment-406536200

    def get_cache_dir():
        return str(cache_dir)

    mp = MonkeyPatch()
    mp.setattr(lektor.project, "get_cache_dir", get_cache_dir)
    yield cache_dir
    mp.undo()


@contextmanager
def restore_import_state():
    """Save `sys.path`, and `sys.modules` state on test
    entry, restore after test completion.

    Any test which constructs a `lektor.environment.Environment` instance
    or which runs any of the Lektor CLI commands should use this fixture
    to ensure that alterations made to `sys.path` do not interfere with
    other tests.

    Lektor's private package cache is added to `sys.path` by
    `lektor.packages.load_packages`.  This happens, for example,
    whenever a Lektor `Environment` is constructed (unless
    `load_plugins=False` is specified.)  Since all tests are run
    within an single invocation of the python interpreter, this can
    cause problems when different tests are using different private
    package caches.

    """
    path = sys.path.copy()
    meta_path = sys.meta_path.copy()
    path_hooks = sys.path_hooks.copy()
    modules = sys.modules.copy()

    # Importlib_metadata, when it is imported, cripples the stdlib distribution finder
    # by deleting its find_distributions method.
    #
    # https://github.com/python/importlib_metadata/blob/705a7571ec7c5abec4d4b008da3a58df7e5c94e7/importlib_metadata/_compat.py#L31
    #
    def clone_class(cls):
        return type(cls)(cls.__name__, cls.__bases__, cls.__dict__.copy())

    sys.meta_path[:] = [
        clone_class(finder) if isinstance(finder, type) else finder
        for finder in meta_path
    ]

    try:
        yield
    finally:
        importlib.invalidate_caches()

        # NB: Restore sys.modules, sys.path, et. all. in place. (Some modules may hold
        # references to these — e.g. pickle appears to hold a reference to sys.modules.)
        for module in set(sys.modules).difference(modules):
            del sys.modules[module]
        sys.modules.update(modules)
        sys.path[:] = path
        sys.meta_path[:] = meta_path
        sys.path_hooks[:] = path_hooks
        sys.path_importer_cache.clear()


_initial_path_key = object()


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_setup(item):
    item.stash[_initial_path_key] = sys.path.copy()


@pytest.hookimpl(trylast=True)
def pytest_runtest_teardown(item):
    # Check that tests don't alter sys.path
    initial_path = item.stash[_initial_path_key]
    assert sys.path == initial_path


@pytest.fixture
def save_sys_path():
    with restore_import_state():
        yield


# Fix for spurious failures in pytest_runtest_teardown sanity-check:
#
# Import setuptools early to pre-mangle sys.path.  This avoids mangling of sys.path
# during a test run, if setuptools is imported during the run.
#
# Details:
# - Setuptools>=71 adds its _vendor subdirectory to sys.path when imported.
# - When running under python < 3.12, setuptools can be imported during execution of
#   import statements.  (This appears to happen, somehow, through setuptools'
#   _distutils_hack:DistUtilsMetafinder:find_spec.)
with suppress(ModuleNotFoundError):
    __import__("setuptools")


@pytest.fixture(scope="session")
def project(data_path):
    return Project.from_path(data_path / "demo-project")


@pytest.fixture(scope="function")
def scratch_project_data(tmp_path):
    base = tmp_path / "scratch-proj"

    def write_text(path, text):
        filename = base / path
        filename.parent.mkdir(parents=True, exist_ok=True)
        filename.write_text(textwrap.dedent(text), "utf-8")

    write_text(
        "Scratch.lektorproject",
        """
        [project]
        name = Scratch

        [alternatives.en]
        primary = yes
        [alternatives.de]
        url_prefix = /de/
        """,
    )
    write_text(
        "content/contents.lr",
        """
        _model: page
        ---
        title: Index
        ---
        body: *Hello World!*
        """,
    )
    write_text(
        "templates/page.html",
        """
        <h1>{{ this.title }}</h1>
        {{ this.body }}
        """,
    )
    write_text(
        "models/page.ini",
        """
        [model]
        label = {{ this.title }}

        [fields.title]
        type = string
        [fields.body]
        type = markdown
        """,
    )

    return base


@pytest.fixture(scope="function")
def scratch_project(scratch_project_data):
    return Project.from_path(scratch_project_data)


@pytest.fixture(scope="function")
def env(project, save_sys_path):
    return Environment(project)


@pytest.fixture(scope="function")
def scratch_env(scratch_project, save_sys_path):
    return Environment(scratch_project)


@pytest.fixture(scope="function")
def pad(env):
    return Database(env).new_pad()


@pytest.fixture(scope="function")
def scratch_pad(scratch_env):
    return Database(scratch_env).new_pad()


@pytest.fixture(scope="function")
def scratch_tree(scratch_pad):
    return Tree(scratch_pad)


@pytest.fixture(scope="function")
def builder(tmp_path, pad):
    output_path = tmp_path / "output"
    output_path.mkdir()
    return Builder(pad, str(output_path))


@pytest.fixture(scope="session")
def built_demo(tmp_path_factory, project):
    output_path = tmp_path_factory.mktemp("demo-output")
    with restore_import_state():
        env = Environment(project)
        builder = Builder(env.new_pad(), os.fspath(output_path))
        builder.build_all()
    return output_path


@pytest.fixture(scope="function")
def scratch_builder(tmp_path, scratch_pad):
    output_path = tmp_path / "output"
    output_path.mkdir()
    return Builder(scratch_pad, str(output_path))


# Builder for child-sources-test-project, a project to test that child sources
# are built even if they're filtered out by a pagination query.
@pytest.fixture(scope="function")
def child_sources_test_project_builder(tmp_path, data_path, save_sys_path):
    output_path = tmp_path / "output"
    output_path.mkdir()
    project = Project.from_path(data_path / "child-sources-test-project")
    pad = project.make_env().new_pad()
    return Builder(pad, str(output_path))


@pytest.fixture(scope="function")
def eval_expr(env):
    def eval_expr(expr, **kwargs):
        expr = Expression(env, expr)
        return expr.evaluate(**kwargs)

    return eval_expr


@pytest.fixture(scope="function")
def reporter(request, env):
    reporter = BufferReporter(env)
    reporter.push()
    request.addfinalizer(reporter.pop)
    return reporter


@pytest.fixture(scope="function")
def project_cli_runner(isolated_cli_runner, project, save_sys_path):
    """
    Copy the project files into the isolated file system used by the
    Click test runner.
    """
    for entry in os.listdir(project.tree):
        entry_path = os.path.join(project.tree, entry)
        if os.path.isdir(entry_path):
            shutil.copytree(entry_path, entry)
        else:
            shutil.copy2(entry_path, entry)
    return isolated_cli_runner


@pytest.fixture
def no_utils(monkeypatch):
    """Monkeypatch $PATH to hide any installed external utilities
    (e.g. git, ffmpeg)."""
    monkeypatch.setitem(os.environ, "PATH", "/dev/null")
    locate_executable.cache_clear()
    try:
        yield
    finally:
        locate_executable.cache_clear()



import os
import posixpath
from collections.abc import Callable
from collections.abc import Iterator
from collections.abc import Mapping
from contextvars import ContextVar
from dataclasses import dataclass
from dataclasses import field
from functools import wraps
from typing import Any
from typing import cast
from typing import TypeVar

import click
import marshmallow
import marshmallow_dataclass as mdcls
from flask import Blueprint
from flask import current_app
from flask import jsonify
from flask import make_response
from flask import request
from flask import Response

from lektor.admin.context import get_lektor_context
from lektor.admin.context import LektorContext
from lektor.admin.utils import eventstream
from lektor.constants import PRIMARY_ALT
from lektor.datamodel import DataModel
from lektor.db import Record
from lektor.environment.config import Config
from lektor.environment.config import ServerInfo
from lektor.publisher import publish
from lektor.publisher import PublishError
from lektor.utils import cleanup_path
from lektor.utils import is_valid_id


bp = Blueprint("api", __name__, url_prefix="/admin/api")


LEKTOR_CONFIG: ContextVar[Config] = ContextVar("lektor_config")


@bp.url_value_preprocessor
def pass_lektor_context(endpoint: str | None, values: dict[str, Any] | None) -> None:
    """Pass LektorContext to each view callable in a `ctx` parameter"""
    assert isinstance(values, dict)
    values["ctx"] = get_lektor_context()


class _ServerInfoField(marshmallow.fields.String):
    def _deserialize(
        self,
        value: str,
        attr: str | None,
        data: Mapping[str, Any] | None,
        **kwargs: Any,
    ) -> ServerInfo:
        lektor_config = LEKTOR_CONFIG.get()
        server_id = super()._deserialize(value, attr, data, **kwargs)
        server_info = lektor_config.get_server(server_id)
        if server_info is None:
            raise marshmallow.ValidationError("Invalid server id.")
        return server_info


def _is_valid_path(value: str) -> None:
    if not cleanup_path(value) == value:
        raise marshmallow.ValidationError("Invalid value.")


def _is_valid_alt(value: str) -> None:
    lektor_config = get_lektor_context().config
    if not lektor_config.is_valid_alternative(value):
        raise marshmallow.ValidationError("Invalid alternative.")


# Mark types for special validation
_PathType = mdcls.NewType("_PathType", str, validate=_is_valid_path)
_AltType = mdcls.NewType("_AltType", str, validate=_is_valid_alt)
_BoolType = mdcls.NewType("_BoolType", bool, truthy={1, "1"}, falsy={0, "0"})


class _SchemaBase(marshmallow.Schema):
    TYPE_MAPPING = {ServerInfo: _ServerInfoField}

    class Meta:
        unknown = marshmallow.EXCLUDE


F = TypeVar("F", bound=Callable[..., Any])


def _with_validated(param_type: type) -> Callable[[F], F]:
    """Flask view decorator to validate parameters.

    The validated parameters are placed into the ``validated`` keyword
    arg of the decorated view.

    If the request has a JSON body, the parameters are parsed from that.
    Otherwise, the parameters are parsed from ``request.values``.

    :param param_type: A dataclass which specifies the parameters.
    """
    schema_class = mdcls.class_schema(param_type, base_schema=_SchemaBase)
    schema = schema_class()

    def wrap(f: F) -> F:
        @wraps(f)
        def wrapper(*args: Any, **kwargs: Any) -> Response:
            if (
                request.method in ("POST", "PUT")
                and request.mimetype == "application/json"
            ):
                data = request.get_json() or {}
            else:
                data = request.values
            LEKTOR_CONFIG.set(kwargs["ctx"].config)
            try:
                kwargs["validated"] = schema.load(data)
            except marshmallow.ValidationError as exc:
                error = {
                    "title": "Invalid parameters",
                    "messages": exc.messages,
                }
                return make_response(jsonify(error=error), 400)
            return f(*args, **kwargs)

        # This cast seems necessary, atm.
        # https://github.com/python/mypy/issues/1927
        return cast(F, wrapper)

    return wrap


@dataclass
class _PathAndAlt:
    path: _PathType
    alt: _AltType = PRIMARY_ALT


@bp.route("/pathinfo")
@_with_validated(_PathAndAlt)
def get_path_info(validated: _PathAndAlt, ctx: LektorContext) -> Response:
    """Returns the path segment information for a record."""
    alt = validated.alt
    tree_item = ctx.tree.get(validated.path)
    segments = []

    while tree_item is not None:
        segments.append(
            {
                "id": tree_item.id,
                "path": tree_item.path,
                "label_i18n": tree_item.get_record_label_i18n(alt),
                "exists": tree_item.exists,
                "can_have_children": tree_item.can_have_children,
            }
        )
        tree_item = tree_item.get_parent()

    segments.reverse()
    return jsonify(segments=segments)


@bp.route("/recordinfo")
@_with_validated(_PathAndAlt)
def get_record_info(validated: _PathAndAlt, ctx: LektorContext) -> Response:
    alt = validated.alt
    tree_item = ctx.tree.get(validated.path)

    return jsonify(
        id=tree_item.id,
        path=tree_item.path,
        label_i18n=tree_item.get_record_label_i18n(alt),
        exists=tree_item.exists,
        is_attachment=tree_item.is_attachment,
        attachments=[
            {
                "id": x.id,
                "path": x.path,
                "type": x.attachment_type,
            }
            for x in tree_item.iter_attachments()
        ],
        children=[
            {
                "id": x.id,
                "path": x.path,
                "label": x.id,
                "label_i18n": x.get_record_label_i18n(alt),
                "visible": x.is_visible,
            }
            for x in tree_item.iter_subpages()
        ],
        alts=[
            {
                "alt": _.id,
                "is_primary": _.id == PRIMARY_ALT,
                "primary_overlay": _.is_primary_overlay,
                "name_i18n": _.name_i18n,
                "exists": _.exists,
            }
            for _ in tree_item.alts.values()
        ],
        can_have_children=tree_item.can_have_children,
        can_have_attachments=tree_item.can_have_attachments,
        can_be_deleted=tree_item.can_be_deleted,
    )


@bp.route("/previewinfo")
@_with_validated(_PathAndAlt)
def get_preview_info(validated: _PathAndAlt, ctx: LektorContext) -> Response:
    record = ctx.pad.get(validated.path, alt=validated.alt)
    if record is None:
        return jsonify(exists=False, url=None, is_hidden=True)
    return jsonify(exists=True, url=record.url_path, is_hidden=record.is_hidden)


@dataclass
class _FindParams:
    q: str
    alt: _AltType = PRIMARY_ALT
    lang: str | None = None


@bp.route("/find", methods=["POST"])
@_with_validated(_FindParams)
def find(validated: _FindParams, ctx: LektorContext) -> Response:
    lang = validated.lang or current_app.config.get("lektor.ui_lang", "en")
    return jsonify(
        results=ctx.builder.find_files(validated.q, alt=validated.alt, lang=lang)
    )


@bp.route("/browsefs", methods=["POST"])
@_with_validated(_PathAndAlt)
def browsefs(validated: _PathAndAlt, ctx: LektorContext) -> Response:
    record = ctx.pad.get(validated.path, alt=validated.alt)
    okay = False
    if record is not None:
        if record.is_attachment:
            fn = record.attachment_filename
        else:
            fn = record.source_filename
        if os.path.exists(fn):
            click.launch(fn, locate=True)
            okay = True
    return jsonify(okay=okay)


@dataclass
class _UrlPath:
    url_path: str


@bp.route("/matchurl")
@_with_validated(_UrlPath)
def match_url(validated: _UrlPath, ctx: LektorContext) -> Response:
    """Find the Record that corresponds to a URL.

    This is used by the admin UI to find the db record that corresponds
    to a page when the preview iframe is navigated.
    """
    record = ctx.pad.resolve_url_path(validated.url_path, alt_fallback=False)
    if not isinstance(record, Record):
        return jsonify(exists=False, path=None, alt=None)
    return jsonify(exists=True, path=record["_path"], alt=record["_alt"])


@bp.route("/rawrecord")
@_with_validated(_PathAndAlt)
def get_raw_record(validated: _PathAndAlt, ctx: LektorContext) -> Response:
    ts = ctx.tree.edit(validated.path, alt=validated.alt)
    return jsonify(ts.to_json())


@bp.route("/newrecord")
@_with_validated(_PathAndAlt)
def get_new_record_info(validated: _PathAndAlt, ctx: LektorContext) -> Response:
    pad = ctx.pad
    alt = validated.alt
    tree_item = ctx.tree.get(validated.path)

    def describe_model(model: DataModel) -> dict[str, Any]:
        primary_field = None
        if model.primary_field is not None:
            f = model.field_map.get(model.primary_field)
            if f is not None:
                primary_field = f.to_json(pad)
        return {
            "id": model.id,
            "name": model.name,
            "name_i18n": model.name_i18n,
            "primary_field": primary_field,
        }

    implied_model = tree_item.implied_child_datamodel
    label_i18n = tree_item.get_record_label_i18n(alt)
    return jsonify(
        {
            "label_i18n": label_i18n,
            "label": label_i18n["en"],
            "can_have_children": tree_item.can_have_children,
            "implied_model": implied_model,
            "available_models": {
                k: describe_model(v)
                for k, v in pad.db.datamodels.items()
                if not v.hidden or k == implied_model
            },
        }
    )


@bp.route("/newattachment")
@_with_validated(_PathAndAlt)
def get_new_attachment_info(validated: _PathAndAlt, ctx: LektorContext) -> Response:
    tree_item = ctx.tree.get(validated.path)
    label_i18n = tree_item.get_record_label_i18n(validated.alt)
    return jsonify(
        {
            "can_upload": tree_item.can_have_attachments,
            "label_i18n": label_i18n,
            "label": label_i18n["en"],
        }
    )


@bp.route("/newattachment", methods=["POST"])
@_with_validated(_PathAndAlt)
def upload_new_attachments(validated: _PathAndAlt, ctx: LektorContext) -> Response:
    ts = ctx.tree.edit(validated.path, alt=validated.alt)
    if not ts.exists or ts.is_attachment:
        return jsonify({"bad_upload": True})

    buckets = []
    for file in request.files.getlist("file"):
        stored_filename = ts.add_attachment(file.filename, file)
        buckets.append(
            {"original_filename": file.filename, "stored_filename": stored_filename}
        )

    return jsonify(
        {
            "bad_upload": False,
            "path": validated.path,
            "buckets": buckets,
        }
    )


@dataclass
class _NewRecordParams:
    id: str
    model: str | None
    data: dict[str, str | None]
    path: _PathType
    alt: _AltType = PRIMARY_ALT


@bp.route("/newrecord", methods=["POST"])
@_with_validated(_NewRecordParams)
def add_new_record(validated: _NewRecordParams, ctx: LektorContext) -> Response:
    exists = False

    if not is_valid_id(validated.id):
        return jsonify(valid_id=False, exists=False, path=None)

    path = posixpath.join(validated.path, validated.id)

    ts = ctx.tree.edit(path, datamodel=validated.model, alt=validated.alt)
    with ts:
        if ts.exists:
            exists = True
        else:
            ts.data.update(validated.data)

    return jsonify({"valid_id": True, "exists": exists, "path": path})


@dataclass
class _DeleteRecordParams:
    delete_master: _BoolType
    path: _PathType
    alt: _AltType = PRIMARY_ALT


@bp.route("/deleterecord", methods=["POST"])
@_with_validated(_DeleteRecordParams)
def delete_record(validated: _DeleteRecordParams, ctx: LektorContext) -> Response:
    if validated.path != "/":
        ts = ctx.tree.edit(validated.path, alt=validated.alt)
        with ts:
            ts.delete(delete_master=validated.delete_master)
    return jsonify(okay=True)


@dataclass
class _UpdateRawRecordParams:
    data: dict[str, str | None]
    path: _PathType
    alt: _AltType = PRIMARY_ALT


@bp.route("/rawrecord", methods=["PUT"])
@_with_validated(_UpdateRawRecordParams)
def update_raw_record(
    validated: _UpdateRawRecordParams, ctx: LektorContext
) -> Response:
    ts = ctx.tree.edit(validated.path, alt=validated.alt)
    with ts:
        ts.data.update(validated.data)
    return jsonify(path=ts.path)


@bp.route("/servers")
def get_servers(ctx: LektorContext) -> Response:
    servers = ctx.config.get_servers(public=True)
    return jsonify(
        servers=sorted(
            [x.to_json() for x in servers.values()], key=lambda x: x["name"].lower()
        )
    )


@bp.route("/build", methods=["POST"])
def trigger_build(ctx: LektorContext) -> Response:
    builder = ctx.builder
    builder.build_all()
    builder.prune()
    return jsonify(okay=True)


@bp.route("/clean", methods=["POST"])
def trigger_clean(ctx: LektorContext) -> Response:
    builder = ctx.builder
    builder.prune(all=True)
    builder.touch_site_config()
    return jsonify(okay=True)


@dataclass
class _PublishBuildParams:
    server_info: ServerInfo = field(metadata={"data_key": "server"})


@bp.route("/publish")
@_with_validated(_PublishBuildParams)
def publish_build(validated: _PublishBuildParams, ctx: LektorContext) -> Response:
    @eventstream
    def generator() -> Iterator[dict[str, str]]:
        try:
            event_iter = (
                publish(
                    ctx.env,
                    validated.server_info.target,
                    ctx.output_path,
                    server_info=validated.server_info,
                )
                or ()
            )
            for event in event_iter:
                yield {"msg": event}
        except PublishError as exc:
            yield {"msg": f"Error: {exc}"}

    return generator()


@bp.route("/ping")
def ping(ctx: LektorContext) -> Response:
    return jsonify(project_id=ctx.project_id, okay=True)



from __future__ import annotations

import dataclasses
import mimetypes
import os
import re
from pathlib import Path
from typing import TYPE_CHECKING
from zlib import adler32

from flask import abort
from flask import Blueprint
from flask import current_app
from flask import render_template
from flask import request
from flask import Response
from flask import send_file
from flask import url_for
from werkzeug.exceptions import NotFound
from werkzeug.security import safe_join
from werkzeug.utils import append_slash_redirect

from lektor.admin.context import get_lektor_context
from lektor.admin.context import LektorApp
from lektor.admin.context import LektorContext
from lektor.assets import Asset
from lektor.assets import Directory
from lektor.constants import PRIMARY_ALT
from lektor.db import Record


if TYPE_CHECKING:
    from flask.typing import ResponseReturnValue
    from flask.typing import ResponseValue

    from lektor.builder import Artifact
    from lektor.buildfailures import BuildFailure
    from lektor.sourceobj import SourceObject


bp = Blueprint("serve", __name__)


Filename = str | os.PathLike[str]


@dataclasses.dataclass(frozen=True)
class LivereloadConfig:
    artifactName: str
    eventsUrl: str
    workerJs: str

    @classmethod
    def from_artifact(cls, artifact: Artifact | None) -> LivereloadConfig | None:
        if artifact is not None and "livereload" in current_app.blueprints:
            return cls(
                artifactName=artifact.artifact_name,
                eventsUrl=url_for("livereload.events"),
                workerJs=url_for("static", filename="livereload-worker.js"),
            )
        return None


@dataclasses.dataclass(frozen=True)
class TooldrawerConfig:
    editUrl: str | None = None
    livereloadConfig: LivereloadConfig | None = None

    def __bool__(self) -> bool:
        return self.editUrl is not None or self.livereloadConfig is not None


def _inject_tooldrawer(
    html: bytes, tooldrawer_config: TooldrawerConfig | None
) -> bytes:
    """Add "edit pencil" and "livereload" control  buttons to the text of an HTML
    page."""
    if tooldrawer_config:
        tooldrawer_html = render_template(
            "tooldrawer.html",
            tooldrawer_config=dataclasses.asdict(tooldrawer_config),
            tooldrawer_js=url_for("static", filename="tooldrawer.js"),
        ).encode("utf-8")
        match = re.search(rb"(?i)</\s*head\s*>|\Z", html)
        assert match is not None
        head_end = match.start()
        html = html[:head_end] + tooldrawer_html + html[head_end:]
    return html


def _send_html_for_editing(
    artifact: Artifact, tooldrawer_config: TooldrawerConfig, mimetype: str = "text/html"
) -> ResponseValue:
    """Serve an HTML file, after mangling it to add an "edit pencil" button."""
    try:
        with open(artifact.dst_filename, "rb") as fp:
            html = fp.read()
            st = os.stat(fp.fileno())
    except (FileNotFoundError, IsADirectoryError, PermissionError):
        abort(404)
    html = _inject_tooldrawer(html, tooldrawer_config)
    check = (
        adler32(f"{artifact.dst_filename}\0{hash(tooldrawer_config)}".encode())
        & 0xFFFFFFFF
    )
    resp = Response(html, mimetype=mimetype)
    resp.set_etag(f"{st.st_mtime}-{st.st_size}-{check}")
    return resp


def _deduce_mimetype(filename: Filename) -> str:
    mimetype = mimetypes.guess_type(filename)[0]
    if mimetype is None:
        mimetype = "application/octet-stream"
    return mimetype


def _checked_send_file(
    filename: Filename, mimetype: str | None = None
) -> ResponseValue:
    """Same as flask.send_file, except raises NotFound on file errors."""
    # NB: flask.send_file interprets relative paths relative to
    # current_app.root_path. We don't want that.
    try:
        resp = send_file(os.path.abspath(filename), mimetype=mimetype)
    except (FileNotFoundError, IsADirectoryError, PermissionError):
        abort(404)
    return resp


class HiddenRecordException(NotFound):
    """Exception thrown when a request is made for a hidden page."""

    def __init__(self, source: SourceObject) -> None:
        super().__init__(description=f"Record is hidden: {source!r}")
        self.source = source


class ArtifactServer:
    """Resolve url_path to a Lektor source object, build it, serve the result.

    Redirects to slash-appended path if appropriate.

    Raises NotFound if source object can not be resolved, or if it does not
    produce an artifact.

    """

    def __init__(self, lektor_context: LektorContext) -> None:
        self.lektor_ctx = lektor_context

    def resolve_url_path(self, url_path: str) -> SourceObject:
        """Resolve URL path to a source object.

        Raise NotFound if resolution fails.
        """
        source = self.lektor_ctx.pad.resolve_url_path(url_path, include_invisible=True)
        if source is None:
            abort(404)
        return source

    @staticmethod
    def resolve_directory_index(directory: Directory) -> Asset:
        """Find an index.html (or equivalent) asset for a Directory asset

        Raise NotFound if no index is found.
        """
        for name in "index.html", "index.htm":
            index = directory.resolve_url_path([name])
            if index is not None:
                break
        else:
            abort(404)
        return index

    def build_primary_artifact(
        self, source: SourceObject
    ) -> tuple[Artifact, BuildFailure | None]:
        """Build source object, return primary artifact.

        If the build was successfull, returns a tuple of (artifact, ``None``).

        If the build failed, returns a tuple of (artifact, failure),
        where failure is an instance of ``BuildFailure`` which
        contains information regarding the failure.

        Raises NotFound if no primary artifact is produced by the build process.
        """
        lektor_ctx = self.lektor_ctx
        with lektor_ctx.cli_reporter():
            prog, _ = lektor_ctx.builder.build(source)
        artifact = prog.primary_artifact
        if artifact is None:
            abort(404)
        failure = lektor_ctx.failure_controller.lookup_failure(artifact.artifact_name)
        return artifact, failure

    @staticmethod
    def handle_build_failure(
        failure: BuildFailure, tooldrawer_config: TooldrawerConfig | None = None
    ) -> Response:
        """Format build failure to an HTML response."""
        html = render_template("build-failure.html", **failure.data).encode("utf-8")
        html = _inject_tooldrawer(html, tooldrawer_config)
        return Response(html, mimetype="text/html")

    def get_edit_url(self, source: SourceObject) -> str | None:
        primary_alternative = self.lektor_ctx.config.primary_alternative
        if not isinstance(source, Record):
            # Asset or VirtualSourceObject — not editable
            return None
        record = source.record
        alt = (
            record.alt if record.alt not in (PRIMARY_ALT, primary_alternative) else None
        )
        return url_for("url.edit", path=record.path, alt=alt)

    def serve_artifact(self, url_path: str) -> ResponseValue:
        source = self.resolve_url_path(url_path)

        # If the request path does not end with a slash but we
        # requested a URL that actually wants a trailing slash, we
        # append it.  This is consistent with what apache and nginx do
        # and it ensures our relative urls work.
        if (
            not url_path.endswith("/")
            and source.url_path.endswith("/")
            and source.url_path != "/"
        ):
            return append_slash_redirect(request.environ)

        if source.is_hidden:
            raise HiddenRecordException(source)

        if isinstance(source, Directory):
            # Special case for asset directories: resolve to index.html
            source = self.resolve_directory_index(source)

        artifact, failure = self.build_primary_artifact(source)
        tooldrawer_config = TooldrawerConfig(
            editUrl=self.get_edit_url(source),
            livereloadConfig=LivereloadConfig.from_artifact(artifact),
        )

        # If there was a build failure for the given artifact, we want
        # to render this instead of sending the (most likely missing or
        # corrupted) file.
        if failure is not None:
            return self.handle_build_failure(failure, tooldrawer_config)

        mimetype = _deduce_mimetype(artifact.dst_filename)
        if mimetype == "text/html" and tooldrawer_config:
            return _send_html_for_editing(artifact, tooldrawer_config, mimetype)
        return _checked_send_file(artifact.dst_filename, mimetype=mimetype)


def serve_artifact(path: str) -> ResponseValue:
    lektor_context = get_lektor_context()
    return ArtifactServer(lektor_context).serve_artifact(path)


def serve_file(path: str) -> ResponseValue:
    """Serve file directly from Lektor's output directory."""
    assert isinstance(current_app, LektorApp)
    output_path = current_app.lektor_info.output_path

    safe_path = safe_join("", *(path.strip("/").split("/")))
    if safe_path is None:
        abort(404)

    filename = Path(output_path, safe_path)  # converts safe_path to native path seps
    if filename.is_dir():
        if not path.endswith("/"):
            return append_slash_redirect(request.environ)
        for index in filename / "index.html", filename / "index.htm":
            if index.is_file():
                return _checked_send_file(index, mimetype="text/html")
        abort(404)

    return _checked_send_file(filename, mimetype=_deduce_mimetype(filename.name))


@bp.route("/", defaults={"path": ""})
@bp.route("/<path:path>")
def serve_artifact_or_file(path: str) -> ResponseReturnValue:
    try:
        return serve_artifact(path)
    except HiddenRecordException:
        raise
    except NotFound:
        return serve_file(path)


@bp.errorhandler(404)
def serve_error_page(error: NotFound) -> ResponseReturnValue:
    try:
        return serve_artifact("404.html"), 404
    except NotFound:
        return error



import pytest

from lektor.filecontents import FileContents


def test_FileContents_is_deprecated():
    with pytest.deprecated_call(match=r"FileContents") as warnings:
        FileContents(__file__)
    assert warnings[0].filename == __file__


def test_FileContents_is_a_type():
    assert isinstance(FileContents, type)



from __future__ import annotations

import datetime
import io
from fractions import Fraction
from pathlib import Path

import pytest
from pytest import approx

from lektor.imagetools.exif import _combine_make
from lektor.imagetools.exif import _to_altitude
from lektor.imagetools.exif import _to_degrees
from lektor.imagetools.exif import _to_flash_description
from lektor.imagetools.exif import _to_float
from lektor.imagetools.exif import _to_focal_length
from lektor.imagetools.exif import _to_rational
from lektor.imagetools.exif import _to_string
from lektor.imagetools.exif import read_exif


HERE = Path(__file__).parent
DEMO_PROJECT = HERE / "../demo-project/content"


@pytest.mark.parametrize(
    "make, model, expected",
    [
        ("Frobnotz", "Frobnicator v42", "Frobnotz Frobnicator v42"),
        ("Frobnotz", None, "Frobnotz"),
        (None, "Frobnicator v42", "Frobnicator v42"),
        ("Frobnotz", "Frobnotz Frobnicator v42", "Frobnotz Frobnicator v42"),
        (None, None, ""),
    ],
)
def test_combine_make(make, model, expected):
    assert _combine_make(make, model) == expected


@pytest.mark.parametrize(
    "coerce, value, expected",
    [
        (_to_string, "a b", "a b"),
        (_to_string, "a\xc2\xa0b", "a\xa0b"),
        (_to_string, "a\xa0b", "a\xa0b"),
        #
        (_to_rational, 42, 42),
        (_to_rational, Fraction("22/7"), Fraction("22/7")),
        (_to_rational, (3, 2), Fraction("3/2")),
        #
        (_to_float, 42, 42),
        (_to_float, 1.5, 1.5),
        (_to_float, Fraction("22/7"), approx(3.142857)),
        (_to_float, (7, 4), 1.75),
        #
        (_to_flash_description, 0x41, "Flash fired, red-eye reduction mode"),
        (_to_flash_description, 0x100, "Flash did not fire (256)"),
        (_to_flash_description, 0x101, "Flash fired (257)"),
        (_to_flash_description, -1, "Flash fired (-1)"),
        #
        (_to_focal_length, Fraction("45/2"), "22.5mm"),
        (_to_focal_length, (521, 10), "52.1mm"),
    ],
)
def test_coersion(coerce, value, expected):
    assert coerce(value) == expected


@pytest.mark.parametrize(
    "coords, hemisphere, expected",
    [
        ((Fraction(45), Fraction(15), Fraction(30)), "N", approx(45.2583333)),
        ((Fraction(45), Fraction(61, 2), Fraction(0)), "S", approx(-45.5083333)),
        ((122, 0, 0), "W", -122),
        ((Fraction("45/2"), 0, 0), "E", 22.5),
        (((45, 2), (30, 1), (0, 1)), "N", approx(23)),
    ],
)
def test_to_degrees(coords, hemisphere, expected):
    assert _to_degrees(coords, hemisphere) == expected


@pytest.mark.parametrize(
    "altitude, altitude_ref, expected",
    [
        (Fraction("1234/10"), b"\x00", approx(123.4)),
        (Fraction("1234/10"), None, approx(123.4)),
        ((1234, 10), b"\x00", approx(123.4)),
        (Fraction("123/10"), b"\x01", approx(-12.3)),
    ],
)
def test_to_altitude(altitude, altitude_ref, expected):
    assert _to_altitude(altitude, altitude_ref) == expected


TEST_JPG_EXIF_INFO = {
    "altitude": approx(779.0293),
    "aperture": approx(2.275),
    "artist": None,
    "camera": "Apple iPhone 6",
    "camera_make": "Apple",
    "camera_model": "iPhone 6",
    "copyright": None,
    "created_at": datetime.datetime(2015, 12, 6, 11, 37, 38),
    "exposure_time": "1/33",
    "f": "ƒ/2.2",
    "f_num": approx(2.2),
    "flash_info": "Flash did not fire, compulsory flash mode",
    "focal_length": "4.2mm",
    "focal_length_35mm": "29mm",
    "iso": 160,
    "latitude": approx(46.633833),
    "lens": "Apple iPhone 6 back camera 4.15mm f/2.2",
    "lens_make": "Apple",
    "lens_model": "iPhone 6 back camera 4.15mm f/2.2",
    "longitude": approx(13.404833),
    "location": approx((46.633833, 13.404833)),
    "shutter_speed": "1/33",
    "documentname": "testName",
    "description": "testDescription",
    "is_rotated": True,
}

# This is the default ExifINFO for images without EXIF data
NULL_EXIF_INFO = {
    "altitude": None,
    "aperture": None,
    "artist": None,
    "camera": "",
    "camera_make": None,
    "camera_model": None,
    "copyright": None,
    "created_at": None,
    "description": None,
    "documentname": None,
    "exposure_time": None,
    "f": None,
    "f_num": None,
    "flash_info": None,
    "focal_length": None,
    "focal_length_35mm": None,
    "is_rotated": False,
    "iso": None,
    "latitude": None,
    "lens": "",
    "lens_make": None,
    "lens_model": None,
    "location": None,
    "longitude": None,
    "shutter_speed": None,
}

TEST_IMAGE_EXIF_INFO = {
    DEMO_PROJECT / "test.jpg": TEST_JPG_EXIF_INFO,
    DEMO_PROJECT / "test-sof-last.jpg": TEST_JPG_EXIF_INFO,
    DEMO_PROJECT / "test-progressive.jpg": NULL_EXIF_INFO,
    HERE / "exif-test-2.gif": NULL_EXIF_INFO,
    (HERE / "exif-test-1.jpg"): {
        "altitude": None,
        "aperture": None,
        "artist": "Geoffrey T. Dairiki",
        "camera": "NIKON CORPORATION NIKON D7100",
        "camera_make": "NIKON CORPORATION",
        "camera_model": "NIKON D7100",
        "copyright": "2015",
        "created_at": datetime.datetime(2022, 10, 22, 9, 20, 56),
        "description": None,
        "documentname": None,
        "exposure_time": "1/400",
        "f": "ƒ/4.5",
        "f_num": 4.5,
        "flash_info": "Flash did not fire, compulsory flash mode",
        "focal_length": "86mm",
        "focal_length_35mm": "129mm",
        "is_rotated": False,
        "iso": 1250,
        "latitude": None,
        "lens": "",
        "lens_make": None,
        "lens_model": None,
        "location": None,
        "longitude": None,
        "shutter_speed": None,
    },
    HERE / "exif-test-3.jpg": {"altitude": approx(-85.9)},  # negative altitude
    HERE / "exif-test-4.jpg": {"altitude": approx(85.9)},  # This no GPSAltitudeRef tag:
}


@pytest.mark.parametrize(
    "path, attr, expected",
    [
        (path, attr, expected)
        for path, tags in TEST_IMAGE_EXIF_INFO.items()
        for attr, expected in tags.items()
    ],
)
def test_read_exif_attr(path, attr, expected):
    with path.open("rb") as fp:
        exif = read_exif(fp)
    assert getattr(exif, attr) == expected


def test_read_exif_unrecognized_image():
    exif_info = read_exif(io.BytesIO(b"unrecognized-image"))
    assert not exif_info


@pytest.fixture(
    params=[
        "test.jpg",  # base image, exif-rotated
        "test-sof-last.jpg",  # same image but with SOF marker last
        "test-progressive.jpg",  # with progressive encoding, rotated in place
    ]
)
def demo_test_jpg(request, pad):
    return pad.root.attachments.images.get(request.param)


def test_image_exif_attr(demo_test_jpg):
    name = demo_test_jpg.path.lstrip("/")
    expected = TEST_IMAGE_EXIF_INFO[DEMO_PROJECT / name]

    assert demo_test_jpg.exif.to_dict() == expected
    if expected == NULL_EXIF_INFO:
        assert not demo_test_jpg.exif
    else:
        assert demo_test_jpg.exif



# pylint: disable=import-outside-toplevel
import os
import sys

import click

from lektor.cli_utils import AliasedGroup
from lektor.cli_utils import extraflag
from lektor.cli_utils import pass_context


try:
    from IPython import embed
    from traitlets.config.loader import Config
except ImportError:
    pass  # fallback to normal Python InteractiveConsole


@click.group(cls=AliasedGroup, short_help="Development commands.")
def cli():
    """Development commands for Lektor.

    This provides various development support commands for Lektor.  This is
    primarily useful for Lektor plugin development but also if you want to
    extend Lektor itself.  Additional functionality can be unlocked by
    exporting the `LEKTOR_DEV=1` environment variable.
    """


@cli.command("shell", short_help="Starts a python shell.")
@extraflag
@pass_context
def shell_cmd(ctx, extra_flags):
    """Starts a Python shell in the context of a Lektor project.

    This is particularly useful for debugging plugins and to explore the
    API.  To quit the shell just use `quit()`.  Within the shell various
    utilities are available right from the get-go for you.

    \b
    - `project`: the loaded project as object.
    - `env`: an environment for the loaded project.
    - `pad`: a database pad initialized for the project and environment
      that is ready to use.
    """
    ctx.load_plugins(extra_flags=extra_flags)
    import code

    from lektor.builder import Builder
    from lektor.db import F
    from lektor.db import Tree

    banner = (
        f"Python {sys.version} on {sys.platform}\n"
        f"Lektor Project: {ctx.get_env().root_path}"
    )
    ns = {}
    startup = os.environ.get("PYTHONSTARTUP")
    if startup and os.path.isfile(startup):
        with open(startup, encoding="utf-8") as f:
            eval(compile(f.read(), startup, "exec"), ns)  # pylint: disable=eval-used
    pad = ctx.get_env().new_pad()
    ns.update(
        project=ctx.get_project(),
        env=ctx.get_env(),
        pad=pad,
        tree=Tree(pad),
        config=ctx.get_env().load_config(),
        make_builder=lambda: Builder(
            ctx.get_env().new_pad(), ctx.get_default_output_path()
        ),
        F=F,
    )
    try:
        c = Config()
        c.TerminalInteractiveShell.banner2 = banner
        embed(config=c, user_ns=ns)
    except NameError:  # No IPython
        code.interact(banner=banner, local=ns)


@cli.command("new-plugin", short_help="Creates a new plugin")
@click.option("--path", type=click.Path(), help="The destination path")
@click.argument("plugin_name", required=False)
@pass_context
def new_plugin(ctx, **defaults):
    """This command creates a new plugin.

    This will present you with a very short wizard that guides you through
    creation of a new plugin.  At the end of it, it will create a plugin
    in the packages folder of the current project or the path you defined.

    This is the fastest way to creating a new plugin.
    """
    from .quickstart import plugin_quickstart

    project = ctx.get_project(silent=True)
    plugin_quickstart(defaults, project=project)


@cli.command("new-theme", short_help="Creates a new theme")
@click.option("--path", type=click.Path(), help="The destination path")
@click.argument("theme_name", required=False)
@pass_context
def new_theme(ctx, **defaults):
    """This command creates a new theme.

    This will present you with a very short wizard that guides you through
    creation of a new theme. At the end of it, it will create a theme
    in the packages folder of the current project or the path you defined.

    This is the fastest way to creating a new theme.
    """
    from .quickstart import theme_quickstart

    project = ctx.get_project(silent=True)
    theme_quickstart(defaults, project=project)



import inspect

import pytest

from lektor.environment.config import Config


def test_custom_attachment_types(env):
    attachment_types = env.load_config().values["ATTACHMENT_TYPES"]
    assert attachment_types[".foo"] == "text"


@pytest.fixture(scope="function")
def config(tmp_path, project_url):
    projectfile = tmp_path / "scratch.lektorproject"
    projectfile.write_text(
        inspect.cleandoc(
            f"""
            [project]
            url = {project_url}
            """
        )
    )
    return Config(projectfile)


@pytest.mark.parametrize(
    "project_url, expected",
    [
        ("", None),
        ("/path/", None),
        ("https://example.org", "https://example.org/"),
    ],
)
def test_base_url(config, expected):
    assert config.base_url == expected


@pytest.mark.parametrize(
    "project_url, expected",
    [
        ("", "/"),
        ("/path", "/path/"),
        ("/path/", "/path/"),
        ("https://example.org", "/"),
        ("https://example.org/pth", "/pth/"),
    ],
)
def test_base_path(config, expected):
    assert config.base_path == expected



from __future__ import annotations

import dataclasses
import io
from pathlib import Path
from xml.sax import saxutils

import PIL
import pytest

from lektor.imagetools.image_info import _parse_svg_units_px
from lektor.imagetools.image_info import _save_position
from lektor.imagetools.image_info import get_image_info


EXIF_ORIENTATION_TAG = 0x0112


@pytest.mark.parametrize(
    "dimension, pixels",
    [
        ("123", 123),
        ("50%", None),
        ("42px", 42),
        ("42.5 px", 42.5),
    ],
)
def test_parse_svg_units_px(dimension, pixels):
    assert _parse_svg_units_px(dimension) == pixels


class DummyFileBase:
    def write(self, fp):
        raise NotImplementedError()

    def open(self):
        fp = io.BytesIO()
        self.write(fp)
        fp.seek(0)
        return fp

    def path(self, tmp_path: Path) -> Path:
        path = tmp_path / "dummy"
        with path.open("wb") as fp:
            self.write(fp)
        return path


@dataclasses.dataclass
class DummyImage(DummyFileBase):
    width: int
    height: int
    format: str = "PNG"
    orientation: int | None = None

    def write(self, fp):
        image = PIL.Image.new("RGB", (self.width, self.height), "#999")
        exif = image.getexif()
        if self.orientation is not None:
            exif[EXIF_ORIENTATION_TAG] = self.orientation
        image.save(fp, self.format, exif=exif)


@dataclasses.dataclass
class DummySVGImage(DummyFileBase):
    width: int | str | None = None
    height: int | str | None = None
    xmlns: str | None = "http://www.w3.org/2000/svg"
    xml_decl: str | None = '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n'

    def write(self, fp):
        values = dataclasses.asdict(self)
        svg_attrs = "".join(
            f" {key}={saxutils.quoteattr(str(values[key]))}"
            for key in ("width", "height", "xmlns")
            if values[key] is not None
        )
        svg = (
            f"{self.xml_decl or ''}"
            f"<svg{svg_attrs}>"
            '<circle cx="50" cy="50" r="50" fill="lightgrey"/>'
            '<circle cx="30" cy="30" r="10" fill="black"/>'
            '<circle cx="60" cy="30" r="10" fill="black"/>'
            '<path stroke="black" d="M30 70 l30 0" stroke-width="5"/>'
            "</svg>"
        )
        fp.write(svg.encode("utf-8"))


@dataclasses.dataclass
class DummyFile(DummyFileBase):
    content: bytes = b""

    def write(self, fp):
        fp.write(self.content)


@pytest.mark.parametrize(
    "image, expected",
    [
        (DummyFile(b"short"), (None, None, None)),
        (DummyFile(b"Not an image. " * 40), (None, None, None)),
        (DummyImage(10, 20, "JPEG"), ("jpeg", 10, 20)),
        (DummyImage(10, 20, "GIF"), ("gif", 10, 20)),
        (DummyImage(10, 20, "PNG"), ("png", 10, 20)),
        (DummyImage(10, 20, "JPEG", orientation=5), ("jpeg", 20, 10)),
        # check that Exif orientation is ignored for PNG images"
        (DummyImage(10, 20, "PNG", orientation=7), ("png", 10, 20)),
        (DummyImage(10, 20, "PPM"), (None, None, None)),
        (DummySVGImage("10px", "20px"), ("svg", 10, 20)),
        (DummySVGImage("10", "20", xml_decl=None), ("svg", 10, 20)),
        (DummySVGImage(None, None), ("svg", None, None)),
        (DummySVGImage("invalid-width", "invalid-height"), ("svg", None, None)),
        (DummySVGImage("10px", "10px", xmlns=None), (None, None, None)),
    ],
)
def test_get_image_info(image, expected, tmp_path):
    image_path = tmp_path / "test-image"
    with image_path.open("wb") as fp:
        image.write(fp)
    assert get_image_info(image_path) == expected


def test_get_image_info_fp_deprecated():
    image = DummyImage(10, 20, "JPEG")
    with pytest.deprecated_call() as w:
        assert get_image_info(image.open()) == ("jpeg", 10, 20)
    for warning in w:
        assert warning.filename == __file__


def test_save_position():
    fp = io.BytesIO(b"data")
    with _save_position(fp) as infp:
        assert infp.read() == b"data"
    assert fp.read() == b"data"


@pytest.fixture(
    params=[
        "test.jpg",  # base image, exif-rotated
        "test-sof-last.jpg",  # same image but with SOF marker last
        "test-progressive.jpg",  # with progressive encoding, rotated in place
    ]
)
def demo_test_jpg(request, pad):
    return pad.root.attachments.images.get(request.param)


def test_image_attributes(demo_test_jpg):
    assert demo_test_jpg.width == 384
    assert demo_test_jpg.height == 512
    assert demo_test_jpg.format == "jpeg"



"""Helper to probe basic image information: dimensions and format."""

from __future__ import annotations

import enum
import re
import warnings
from collections.abc import Generator
from collections.abc import Mapping
from contextlib import contextmanager
from contextlib import ExitStack
from contextlib import suppress
from pathlib import Path
from typing import BinaryIO
from typing import Final
from typing import NamedTuple
from typing import TYPE_CHECKING
from typing import TypeAlias
from xml.etree import ElementTree as etree

import PIL.Image

from ._compat import ExifTags
from ._compat import UnidentifiedImageError


if TYPE_CHECKING:
    from _typeshed import SupportsRead
    from typing import Literal


class SvgImageInfo(NamedTuple):
    format: Literal["svg"] = "svg"
    width: float | None = None
    height: float | None = None


class PILImageInfo(NamedTuple):
    format: str
    width: int
    height: int


class UnknownImageInfo(NamedTuple):
    format: None = None
    width: None = None
    height: None = None


ImageInfo: TypeAlias = PILImageInfo | SvgImageInfo | UnknownImageInfo


class TiffOrientation(enum.IntEnum):
    """The possible values of the "Exif Orientation" tag."""

    TOPLEFT = 1
    TOPRIGHT = 2
    BOTRIGHT = 3
    BOTLEFT = 4
    LEFTTOP = 5
    RIGHTTOP = 6
    RIGHTBOT = 7
    LEFTBOT = 8

    def __init__(self, value: int):
        # True if orientation implies width and height are transposed
        self.is_transposed = value in {5, 6, 7, 8}


def get_image_orientation(image: PIL.Image.Image) -> TiffOrientation:
    """Deduce the orientation of the image.

    Notes
    -----

    Note that browsers only seem to respect the "Exif" Orientation tag for JPEG images
    (and probably TIFF images). In particular, it is `typically ignored`__ by browsers
    when displaying PNG, WEBP and AVIF files (though AVIF files have their own way of
    indicating orientation — the "irot" and "imir" properties — which are respected.)

    __ https://zpl.fi/exif-orientation-in-different-formats/

    Exif information can be stored in PNG files, however the `spec for Exif in PNG`__
    does state that the Exif information should be considered "historical", under the
    assumption that it was probably copied directly from the source image (where it was
    written by, e.g., the camera). It implies that the "unsafe-to-copy" Exif information
    (e.g. orientation, size) should be ignored.

    __ https://ftp-osl.osuosl.org/pub/libpng/documents/proposals/eXIf/png-proposed-eXIf-chunk-2017-06-15.html

    Prior to Lektor 3.4, Lektor's ``get_image_info`` only checked the orientation for
    JPEG images (transposing width↔height when appropriate).  The ``-auto-orient``
    option to ImageMagick's ``convert`` appears to ignore Exif Orientation in PNG files,
    too.

    Finally, note that reading Exif information from PNG files using Pillow is a slow
    operation.  It seems to require loading and decoding the full image.  (Loading Exif
    information from JPEG files does not require decoding the image, so is much
    quicker.)

    For all of these reasons, we only check the Exif Orientation tag for JPEGs.

    """  # pylint: disable=line-too-long # noqa: E501
    if image.format != "JPEG":
        return TiffOrientation.TOPLEFT
    exif = image.getexif()
    try:
        orientation = exif[ExifTags.Base.Orientation]
        return TiffOrientation(orientation)
    except (ValueError, LookupError):
        return TiffOrientation.TOPLEFT


def _parse_svg_units_px(length: str) -> float | None:
    match = re.match(
        r"\d+(?: \.\d* )? (?= (?: \s*px )? \Z)", length.strip(), re.VERBOSE
    )
    if match:
        return float(match.group())
    return None


class BadSvgFile(Exception):
    """Exception raised when SVG file can not be parsed."""


def _get_svg_info(
    source: str | Path | SupportsRead[bytes],
) -> SvgImageInfo | UnknownImageInfo:
    try:
        _, svg = next(etree.iterparse(source, events=["start"]))
    except (etree.ParseError, StopIteration) as exc:
        raise BadSvgFile("can not parse SVG file") from exc
    if svg.tag != "{http://www.w3.org/2000/svg}svg":
        raise BadSvgFile("unknown tag in SVG file")
    width = _parse_svg_units_px(svg.attrib.get("width", ""))
    height = _parse_svg_units_px(svg.attrib.get("height", ""))
    return SvgImageInfo("svg", width, height)


# Mapping from PIL format to Lektor format
_LEKTOR_FORMATS: Final[Mapping[str, str]] = {
    "PNG": "png",
    "GIF": "gif",
    "JPEG": "jpeg",
}


def _PIL_image_info(
    image: PIL.Image.Image,
) -> PILImageInfo | UnknownImageInfo:
    """Determine image format and dimensions for PIL Image"""

    assert image.format is not None
    try:
        lektor_fmt = _LEKTOR_FORMATS[image.format]
    except LookupError:
        return UnknownImageInfo()

    width = image.width
    height = image.height

    orientation = get_image_orientation(image)
    if orientation.is_transposed:
        width, height = height, width

    return PILImageInfo(lektor_fmt, width, height)


@contextmanager
def _save_position(fp: BinaryIO) -> Generator[BinaryIO]:
    position = fp.tell()
    try:
        yield fp
    finally:
        fp.seek(position)


def get_image_info(source: str | Path | BinaryIO) -> ImageInfo:
    """Determine type and dimensions of an image file."""
    with suppress(UnidentifiedImageError), ExitStack() as stack:
        if not isinstance(source, (str, Path)):
            warnings.warn(
                "Passing a file object to 'get_image_info' is deprecated "
                "since version 3.4.0. Pass a file path instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            stack.enter_context(_save_position(source))

        image = stack.enter_context(PIL.Image.open(source))
        return _PIL_image_info(image)

    with suppress(BadSvgFile), ExitStack() as stack:
        if not isinstance(source, (str, Path)):
            stack.enter_context(_save_position(source))
        return _get_svg_info(source)

    return UnknownImageInfo()
